# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import copy
import warnings

from synalinks.src import ops
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import Embedding as EmbeddingVector
from synalinks.src.backend import JsonDataModel
from synalinks.src.backend import is_entities
from synalinks.src.backend import is_entity
from synalinks.src.backend import is_knowledge_graph
from synalinks.src.backend import is_relation
from synalinks.src.backend import is_relations
from synalinks.src.modules.module import Module
from synalinks.src.saving import serialization_lib


@synalinks_export(
    [
        "synalinks.modules.Embedding",
        "synalinks.Embedding",
    ]
)
class Embedding(Module):
    """Extracts and updates the embedding vector of entities.

    This module is designed to work with `Entity`, `Relation`, `Entities`,
    `Relations` or `KnowledgeGraph` data models. It supports to mask the
    entity fields in order to keep **only one** field to embed per entity.

    **Note**: Each entity should have the *same field* to compute the embedding
        from like a `name` or `description` field using `in_mask`.
        **Or** every entity should have *only one field left* after masking using
        `out_mask` argument.

    ```python
    import synalinks
    import asyncio
    from typing import Literal

    class Document(synalinks.Entity):
        label: Literal["Document"]
        text: str = synalinks.Field(
            description="The document content",
        )

    async def main():
        inputs = synalinks.Input(data_model=Document)
        outputs = await synalinks.Embedding(
            embedding_model=embedding_model,
            in_mask=["text"],
        )(inputs)

        program = synalinks.Program(
            inputs=inputs,
            outputs=outputs,
            name="embbed_document",
            description="Embbed the given documents"
        )

        doc = Document(
            label="Document",
            text="my document",
        )

        result = await program(doc)

    if __name__ == "__main__":
        asyncio.run(main())
    ```

    If you want to process batch asynchronously
    use `program.predict()` instead, see the [FAQ](https://synalinks.github.io/synalinks/FAQ/#whats-the-difference-between-program-methods-predict-and-__call__)
    to understand the difference between `program()` and `program.predict()`

    Here is an example:

    ```python
    import synalinks
    import asyncio
    import numpy as np
    from typing import Literal

    class Document(synalinks.Entity):
        label: Literal["Document"]
        text: str = synalinks.Field(
            description="The document content",
        )

    async def main():
        inputs = synalinks.Input(data_model=Document)
        outputs = await synalinks.Embedding(
            embedding_model=embedding_model,
            in_mask=["text"],
        )(inputs)

        program = synalinks.Program(
            inputs=inputs,
            outputs=outputs,
            name="embbed_document",
            description="Embbed the given documents"
        )

        doc1 = Document(label="Document", text="my document 1")
        doc2 = Document(label="Document", text="my document 2")
        doc3 = Document(label="Document", text="my document 3")

        docs = np.array([doc1, doc2, doc3], dtype="object")

        embedded_docs = await program.predict(docs)

    if __name__ == "__main__":
        asyncio.run(main())
    ```

    Args:
        embedding_model (EmbeddingModel): The embedding model to use.
        in_mask (list): A mask applied to keep specific entity fields.
        out_mask (list): A mask applied to remove specific entity fields.
        name (str): Optional. The name of the module.
        description (str): Optional. The description of the module.
        trainable (bool): Whether the module's variables should be trainable.
    """

    def __init__(
        self,
        embedding_model=None,
        in_mask=None,
        out_mask=None,
        name=None,
        description=None,
        trainable=False,
    ):
        super().__init__(
            name=name,
            description=description,
            trainable=trainable,
        )
        self.embedding_model = embedding_model
        self.in_mask = in_mask
        self.out_mask = out_mask

    async def _embed_entity(self, entity):
        # # Check if entity is already embedded and has valid embeddings
        embeddings = entity.get("embeddings")
        if embeddings:
            warnings.warn(
                "Embeddings already generated for entity.Returning original entity."
            )
            return JsonDataModel(
                json=entity.get_json(),
                schema=entity.get_schema(),
                name=entity.name + "_embedded",
            )
        # Apply masking to the entity
        filtered_entity = entity  # Default to original entity

        if self.out_mask:
            filtered_entity = await ops.out_mask(
                entity,
                mask=self.out_mask,
                recursive=False,
                name=entity.name + "_out_mask",
            )
        elif self.in_mask:
            filtered_entity = await ops.in_mask(
                entity,
                mask=self.in_mask,
                recursive=False,
                name=entity.name + "_in_mask",
            )

        # Generate embeddings
        embeddings = await ops.embedding(
            filtered_entity,
            embedding_model=self.embedding_model,
            name=entity.name + "_embedding",
        )

        # Validate embeddings
        if not embeddings or not embeddings.get("embeddings"):
            warnings.warn(
                f"No embeddings generated for entity {entity.name}. "
                "Please check that your schema is correct."
            )
            return None

        embedding_list = embeddings.get("embeddings")
        if len(embedding_list) != 1:
            warnings.warn(
                "Entities can only have one embedding vector per entity, "
                "adjust `Embedding` module's `in_mask` or `out_mask` "
                "to keep only one field. Skipping embedding."
            )
            return None

        # Add embedding to entity
        vector = embedding_list[0]
        return await ops.concat(
            entity,
            EmbeddingVector(embedding=vector),
            name=entity.name + "_embedded",
        )

    async def _embed_relation(self, relation):
        subj = relation.get_nested_entity("subj")
        obj = relation.get_nested_entity("obj")
        if not subj or not obj:
            return None
        embedded_subj = await self._embed_entity(subj)
        embedded_obj = await self._embed_entity(obj)

        relation_json = copy.deepcopy(relation.get_json())
        relation_json.update(
            {
                "subj": embedded_subj.get_json(),
                "obj": embedded_obj.get_json(),
            }
        )
        outputs_schema = copy.deepcopy(relation.get_schema())

        # Update schema definitions for embedded entities
        if outputs_schema.get("$defs"):
            subj_label = subj.get("label")
            obj_label = obj.get("label")

            if subj_label and subj_label in outputs_schema["$defs"]:
                embedded_subj_schema = embedded_subj.get_schema()
                if embedded_subj_schema.get("properties"):
                    outputs_schema["$defs"][subj_label]["properties"].update(
                        embedded_subj_schema["properties"]
                    )

            if obj_label and obj_label in outputs_schema["$defs"]:
                embedded_obj_schema = embedded_obj.get_schema()
                if embedded_obj_schema.get("properties"):
                    outputs_schema["$defs"][obj_label]["properties"].update(
                        embedded_obj_schema["properties"]
                    )

        return JsonDataModel(
            json=relation_json,
            schema=outputs_schema,
            name=relation.name + "_embedded",
        )

    async def call(self, inputs):
        if not inputs:
            return None
        if is_knowledge_graph(inputs):
            entities_json = []
            relations_json = []
            outputs_schema = copy.deepcopy(inputs.get_schema())

            # Process entities
            for entity in inputs.get_nested_entity_list("entities"):
                embedded_entity = await self._embed_entity(entity)
                if embedded_entity:
                    entities_json.append(embedded_entity.get_json())

                    # Update schema definitions
                    if outputs_schema.get("$defs"):
                        entity_label = entity.get("label")
                        if entity_label and entity_label in outputs_schema["$defs"]:
                            embedded_schema = embedded_entity.get_schema()
                            if embedded_schema.get("properties"):
                                outputs_schema["$defs"][entity_label][
                                    "properties"
                                ].update(embedded_schema["properties"])
            # Process relations
            for relation in inputs.get_nested_entity_list("relations"):
                embedded_relation = await self._embed_relation(relation)
                if embedded_relation:
                    relations_json.append(embedded_relation.get_json())

                    embedded_schema = embedded_relation.get_schema()
                    if embedded_schema.get("$defs") and outputs_schema.get("$defs"):
                        for def_key, def_value in embedded_schema["$defs"].items():
                            if def_key in outputs_schema["$defs"]:
                                # Merge properties if they exist
                                if def_value.get("properties") and outputs_schema[
                                    "$defs"
                                ][def_key].get("properties"):
                                    outputs_schema["$defs"][def_key]["properties"].update(
                                        def_value["properties"]
                                    )
                                else:
                                    outputs_schema["$defs"][def_key] = def_value

            # Update output JSON
            outputs_json = inputs.get_json()
            outputs_json.update({"entities": entities_json, "relations": relations_json})
            return JsonDataModel(
                json=outputs_json,
                schema=outputs_schema,
                name=inputs.name + "_embedded",
            )

        elif is_entities(inputs):
            entities_json = []
            outputs_schema = copy.deepcopy(inputs.get_schema())

            # Process all entities and collect schema updates
            for entity in inputs.get_nested_entity_list("entities"):
                embedded_entity = await self._embed_entity(entity)
                if embedded_entity:
                    entities_json.append(embedded_entity.get_json())

                    # Update schema definitions
                    if outputs_schema.get("$defs"):
                        entity_label = entity.get("label")
                        if entity_label and entity_label in outputs_schema["$defs"]:
                            embedded_schema = embedded_entity.get_schema()
                            if embedded_schema.get("properties"):
                                outputs_schema["$defs"][entity_label][
                                    "properties"
                                ].update(embedded_schema["properties"])

            # Update output JSON with embedded entities
            outputs_json = inputs.get_json()
            outputs_json.update({"entities": entities_json})

            return JsonDataModel(
                json=outputs_json,
                schema=outputs_schema,
                name=inputs.name + "_embedded",
            )

        elif is_relations(inputs):
            relations_json = []
            outputs_schema = copy.deepcopy(inputs.get_schema())

            # Process all relations
            for relation in inputs.get_nested_entity_list("relations"):
                embedded_relation = await self._embed_relation(relation)
                if embedded_relation:
                    relations_json.append(embedded_relation.get_json())

                    # Merge schema definitions from embedded relation
                    embedded_schema = embedded_relation.get_schema()
                    if embedded_schema.get("$defs") and outputs_schema.get("$defs"):
                        for def_key, def_value in embedded_schema["$defs"].items():
                            if def_key in outputs_schema["$defs"]:
                                # Merge properties if they exist
                                if def_value.get("properties") and outputs_schema[
                                    "$defs"
                                ][def_key].get("properties"):
                                    outputs_schema["$defs"][def_key]["properties"].update(
                                        def_value["properties"]
                                    )
                                else:
                                    outputs_schema["$defs"][def_key] = def_value

            # Update output JSON
            outputs_json = inputs.get_json()
            outputs_json.update({"relations": relations_json})

            return JsonDataModel(
                json=outputs_json,
                schema=outputs_schema,
                name=inputs.name + "_embedded",
            )
        elif is_relation(inputs):
            return await self._embed_relation(inputs)
        elif is_entity(inputs):
            return await self._embed_entity(inputs)
        else:
            return None

    async def compute_output_spec(self, inputs):
        return inputs.clone(name=inputs.name + "_embedded")

    def get_config(self):
        config = {
            "in_mask": self.in_mask,
            "out_mask": self.out_mask,
            "name": self.name,
            "description": self.description,
            "trainable": self.trainable,
        }
        embedding_model_config = {
            "embedding_model": serialization_lib.serialize_synalinks_object(
                self.embedding_model
            )
        }
        return {**embedding_model_config, **config}

    @classmethod
    def from_config(cls, config):
        embedding_model = serialization_lib.deserialize_synalinks_object(
            config.pop("embedding_model")
        )
        return cls(embedding_model=embedding_model, **config)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from typing import Any
from typing import List
from typing import Optional

from synalinks.src import tree
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import DataModel
from synalinks.src.backend import Field
from synalinks.src.backend import Trainable
from synalinks.src.backend import out_mask_json
from synalinks.src.backend.common import numpy as np
from synalinks.src.modules.core.input_module import Input
from synalinks.src.modules.ttc.chain_of_thought import ChainOfThought
from synalinks.src.optimizers.random_few_shot import RandomFewShot
from synalinks.src.programs.program import Program
from synalinks.src.rewards.reward import squeeze_or_expand_to_same_rank
from synalinks.src.saving import serialization_lib


def base_instructions():
    """
    Base instructions that define the context for all optimization programs.
    These instructions explain that the system optimizes JSON variables in a computation graph.
    """
    return """
You are an integral part of an optimization system designed to improve JSON variables within a computation graph (i.e. the program).
Each module in the graph performs specific computations, with JSON variables serving as the state.
These variables can represent prompts, code, plans, rules, or any other JSON-compatible data.
""".strip()


def mutation_instructions(variables_keys):
    """
    Instructions for the mutation program that optimizes variables.

    Args:
        variables_keys (list): List of keys that the variable should contain
    """
    return f"""
Your primary task is to creatively enhance the provided variable so that the predicted output aligns as closely as possible with the ground truth.
Pay close attention to the variable's description, its intended use, and the broader context of the computation graph.

Guidelines:
- Ensure the new variable is generalizable and performs well across various inputs of the same kind.
- Include all specified keys: {variables_keys}.
- Justify each change with clear reasoning, referencing the variable's purpose and the desired output.
- If no ground truth is provided, the goal is to critically enhance the predicted output.
""".strip()


class MutationInputs(DataModel):
    program_description: str = Field(
        description="The program description",
    )
    program_inputs: List[Any] = Field(
        description="The inputs of the program",
    )
    program_predicted_outputs: List[Any] = Field(
        description="The program's predicted outputs",
    )
    program_ground_truth: List[Optional[Any]] = Field(
        description="The program's ground truth",
    )
    variable_description: str = Field(
        description="The description of the variable to optimize within that program"
    )
    current_variable: Any = Field(
        description="The variable to optimize",
    )


def crossover_instructions(variables_keys):
    """
    Instructions for the crossover program that optimizes variables.

    Args:
        variables_keys (list): List of keys that the variable should contain
    """
    return f"""
Your responsibility is to create a new, optimized variable by strategically combining features from the current variable and a high-performing candidate.
The new variable should improve the alignment of the predicted output with the ground truth.

Guidelines:
- Analyze both the current variable and the other high-performing variable, identifying their respective strengths and weaknesses.
- Pay close attention to the variable's description, its intended use, and the broader context of the computation graph.
- Ensure the new variable is generalizable and performs well across various inputs of the same kind.
- Include all specified keys: {variables_keys}.
- Justify each feature you incorporate, explaining how it contributes to better performance or alignment with the ground truth.
- If no ground truth is provided, the goal is to critically enhance the predicted output. 
""".strip()


class CrossoverInputs(DataModel):
    program_description: str = Field(
        description="The program description",
    )
    program_inputs: List[Any] = Field(
        description="The inputs of the program",
    )
    program_predicted_outputs: List[Any] = Field(
        description="The program's predicted outputs",
    )
    program_ground_truth: List[Optional[Any]] = Field(
        description="The program's ground truth",
    )
    variable_description: str = Field(
        description="The description of the variable to optimize within that program",
    )
    other_variable: Any = Field(
        description="other high performing variable to merge",
    )
    current_variable: Any = Field(
        description="current high performing variable to merge",
    )


async def similarity_distance(candidate1, candidate2, embedding_model=None, axis=-1):
    """The default cosine similarity distance used by Dominated Novelty Search

    If the candidates have multiple fields, they are combined by averaging the
    vectors before calculating the distance allowing candidates to have
    variable/dynamic number of fields.

    Args:
        candidate1 (dict): The first variable candidate
        candidate2 (dict): The second variable candidate
        embedding_model (EmbeddingModel): The embedding model to use
        axis (int): The axis along which compute the similarity (default -1)
    """
    embeddings1 = await embedding_model(tree.flatten(candidate1))
    embeddings2 = await embedding_model(tree.flatten(candidate2))
    embeddings1 = embeddings1["embeddings"]
    embeddings2 = embeddings2["embeddings"]
    embeddings1 = np.convert_to_tensor(embeddings1)
    embeddings2 = np.convert_to_tensor(embeddings2)
    embeddings1, embeddings2 = squeeze_or_expand_to_same_rank(embeddings1, embeddings2)
    embeddings1 = np.normalize(embeddings1, axis=axis)
    embeddings2 = np.normalize(embeddings2, axis=axis)
    embeddings1 = np.mean(embeddings1, axis=0)
    embeddings1 = np.mean(embeddings2, axis=0)
    similarity = (np.sum(embeddings1 * embeddings2, axis=axis) + 1) / 2
    return 1 - similarity


@synalinks_export("synalinks.optimizers.OMEGA")
class OMEGA(RandomFewShot):
    """OMEGA: OptiMizEr as Genetic Algorithm - A genetic optimizer with dominated novelty search.

    This optimizer is **unique to Synalinks** and the result of our research effort on advancing neuro-symbolic AI.

    Dominated Novelty Search (DNS), is a SOTA Quality-Diversity optimization method that implements a competition function in a classic genetic algorithm.

    The key insight behind Dominated Novelty Search is that candidates should be eliminated from the population if they are both:

    - Inferior in reward/fitness
    - Similar to existing candidates/solutions

    This algorithm creates an evolutionary pressure to focus on high performing candidates **Or** candidates that explore other approaches.

    This approach only add one step to the traditional genetic algorithm and *outperform* MAP-Elites, Threshold-Elites and Cluster-Elites.

    This allow the system to explore the search space more quickly by eliminating non-promising candidates while preserving diversity to avoid local optimum.

    At Synalinks, we adapted this algorithm for LM-based optimization, to do so we use an embedding model to compute the candidate's descriptor and a cosine distance between solutions.

    **Note**: In Synalinks, unlike other In-Context learning frameworks, a variable (the module's state to optimize) is a JSON object not a simple string.
    Which has multiple implications, we maintain a 100% correct structure through constrained JSON decoding, and we allow the state to have variable/dynamic
    number of fields, which is handled by this approach by embedding each field and averaging them before computing the distance required by DNS.

    Example:
    ```
    import synalinks
    import asyncio

    async def main():
        # ... your program definition

        program.compile(
            reward=synalinks.rewards.ExactMatch(),
            optimizer=synalinks.optimizers.OMEGA(
                language_model=language_model,
                embedding_model=embedding_model,
            )
        )

        history = await program.fit(...)
    ```

    Concerning the inspirations for this optimizer:
        - Dominated Novelty Search for their elegant Quality-Diversity algorithm that outperform many other evolutionary strategies.
        - DSPY's GEPA for feeding the optimizer program with the raw training data and for formalizing the evolutionary optimization strategy (**not** the MAP-Elites method used).
        - DeepMind's AlphaEvolve have been a huge inspiration, more on the motivational side as they didn't released the code.

    References:
        - [Dominated Novelty Search: Rethinking Local Competition in Quality-Diversity](https://arxiv.org/html/2502.00593v1)
        - [GEPA: Reflective Prompt Evolution Can Outperform Reinforcement Learning](https://arxiv.org/pdf/2507.19457)
        - [AlphaEvolve: A coding agent for scientific and algorithmic discovery](https://arxiv.org/pdf/2506.13131)

    Args:
        language_model (LanguageModel): The language model to use.
        embedding_model (EmbeddingModel): The embedding model to use to compute candidates
            descriptors according to Dominated Novelty Search.
        k_nearest_fitter (int): The K nearest fitter used by Dominated Novelty Search.
        distance_function (callable): Optional. The distance function to use by Dominated Novelty Search.
            If no function is provided, use the default cosine distance.
        mutation_temperature (float): The temperature for the LM calls of the mutation programs.
        crossover_temperature (float): The temperature for the LM calls of the crossover programs.
        few_shot_learning (bool): If `True` enable the selection of examples using
            the same method than the `RandomFewShot` optimizer.
        nb_min_examples (int): The min number of examples for few-shot learning (Default to 1).
        nb_max_examples (int): The max number of examples for few-shot learning (Default to 3).
        sampling_temperature (float): The temperature for softmax sampling of the few-shot
            learning examples. Lower values concentrate sampling on high-reward predictions,
            higher values make sampling more uniform (Default 1.0).
        merging_rate (float): Rate at which crossover vs mutation is selected. (Default to 0.02).
        population_size (int): The maximum number of best candidates to keep
            during the optimization process.
        name (str): Optional name for the optimizer instance.
        description (str): Optional description of the optimizer instance.
    """

    def __init__(
        self,
        language_model=None,
        embedding_model=None,
        k_nearest_fitter=5,
        distance_function=None,
        mutation_temperature=1.0,
        crossover_temperature=1.0,
        merging_rate=0.02,
        few_shot_learning=False,
        nb_min_examples=1,
        nb_max_examples=3,
        sampling_temperature=1.0,
        population_size=10,
        name=None,
        description=None,
        **kwargs,
    ):
        super().__init__(
            nb_min_examples=nb_min_examples,
            nb_max_examples=nb_max_examples,
            sampling_temperature=sampling_temperature,
            merging_rate=merging_rate,
            population_size=population_size,
            name=name,
            description=description,
        )
        self.language_model = language_model
        self.embedding_model = embedding_model
        self.mutation_temperature = mutation_temperature
        self.crossover_temperature = crossover_temperature
        self.k_nearest_fitter = k_nearest_fitter
        self.few_shot_learning = few_shot_learning

        if not distance_function:
            self.distance_function = similarity_distance
        else:
            self.distance_function = distance_function

        self.kwargs = kwargs

        self.mutation_programs = {}
        self.crossover_programs = {}

    async def build(self, trainable_variables):
        """
        Build the optimizer programs based on the trainable variables.

        Args:
            trainable_variables (list): List of variables that will be optimized
        """
        for trainable_variable in trainable_variables:
            schema_id = id(trainable_variable.get_schema())
            mask = list(Trainable.keys())
            symbolic_variable = trainable_variable.to_symbolic_data_model().out_mask(
                mask=mask
            )

            if schema_id not in self.mutation_programs:
                inputs = Input(data_model=MutationInputs)
                outputs = await ChainOfThought(
                    data_model=symbolic_variable,
                    language_model=self.language_model,
                    temperature=self.mutation_temperature,
                    instructions="\n".join(
                        [
                            base_instructions(),
                            mutation_instructions(list(symbolic_variable.keys())),
                        ]
                    ),
                )(inputs)
                outputs = outputs.in_mask(mask=list(symbolic_variable.keys()))
                program = Program(
                    inputs=inputs,
                    outputs=outputs,
                    name=f"{trainable_variable.name}_mutation",
                    description="The mutation program that fix/optimize variables",
                )
                self.mutation_programs[schema_id] = program

            if schema_id not in self.crossover_programs:
                inputs = Input(data_model=CrossoverInputs)
                outputs = await ChainOfThought(
                    data_model=symbolic_variable,
                    language_model=self.language_model,
                    temperature=self.crossover_temperature,
                    instructions="\n".join(
                        [
                            base_instructions(),
                            crossover_instructions(list(symbolic_variable.keys())),
                        ]
                    ),
                )(inputs)
                outputs = outputs.in_mask(mask=list(symbolic_variable.keys()))
                program = Program(
                    inputs=inputs,
                    outputs=outputs,
                    name=f"{trainable_variable.name}_crossover",
                    description="The crossover program that combine high performing variables",
                )
                self.crossover_programs[schema_id] = program

        self.built = True

    async def propose_new_candidates(
        self,
        step,
        trainable_variables,
        x=None,
        y=None,
        y_pred=None,
        training=False,
    ):
        variable_name_to_update = await self.select_variable_name_to_update(
            trainable_variables,
        )

        strategy = await self.select_evolving_strategy()

        for trainable_variable in trainable_variables:
            if trainable_variable.name == variable_name_to_update:
                mask = list(Trainable.keys())
                schema_id = id(trainable_variable.get_schema())
                if strategy == "mutation":
                    masked_variable = out_mask_json(
                        trainable_variable.get_json(),
                        mask=mask,
                    )
                    inputs = MutationInputs(
                        program_description=self.program.description,
                        program_inputs=[inp.get_json() for inp in x],
                        program_predicted_outputs=[
                            pred.get_json() if pred else None for pred in y_pred
                        ],
                        program_ground_truth=(
                            [gt.get_json() for gt in y] if y is not None else []
                        ),
                        variable_description=trainable_variable.description,
                        current_variable=masked_variable,
                    )
                    program = self.mutation_programs[schema_id]
                    new_candidate = await program(inputs, training=training)
                    if self.few_shot_learning:
                        examples = await self.sample_best_predictions(
                            trainable_variable,
                        )
                    else:
                        examples = None
                elif strategy == "crossover":
                    candidate_to_merge = await self.select_candidate_to_merge(
                        step,
                        trainable_variable,
                    )
                    if candidate_to_merge:
                        current_variable = out_mask_json(
                            trainable_variable.get_json(),
                            mask=mask,
                        )
                        other_variable = out_mask_json(
                            candidate_to_merge,
                            mask=mask,
                        )
                        inputs = CrossoverInputs(
                            program_description=self.program.description,
                            program_inputs=[inp.get_json() for inp in x],
                            program_predicted_outputs=[
                                pred.get_json() if pred else None for pred in y_pred
                            ],
                            program_ground_truth=(
                                [gt.get_json() for gt in y] if y is not None else []
                            ),
                            variable_description=trainable_variable.description,
                            other_variable=other_variable,
                            current_variable=current_variable,
                        )
                        program = self.crossover_programs[schema_id]
                        new_candidate = await program(inputs, training=training)
                        if self.few_shot_learning:
                            examples = self.merge_examples(
                                trainable_variable.get("examples"),
                                candidate_to_merge.get("examples"),
                            )
                        else:
                            examples = None
                    else:
                        masked_variable = out_mask_json(
                            trainable_variable.get_json(),
                            mask=mask,
                        )
                        inputs = MutationInputs(
                            program_description=self.program.description,
                            program_inputs=[inp.get_json() for inp in x],
                            program_predicted_outputs=[
                                pred.get_json() if pred else None for pred in y_pred
                            ],
                            program_ground_truth=(
                                [gt.get_json() for gt in y] if y is not None else []
                            ),
                            variable_description=trainable_variable.description,
                            current_variable=masked_variable,
                        )
                        program = self.mutation_programs[schema_id]
                        new_candidate = await program(inputs, training=training)
                        if self.few_shot_learning:
                            examples = await self.sample_best_predictions(
                                trainable_variable,
                            )
                        else:
                            examples = None

                await self.assign_candidate(
                    trainable_variable,
                    new_candidate=new_candidate,
                    examples=examples,
                )

    async def competition(
        self,
        candidates,
    ):
        """
        This function implement Dominated Novelty Search.

        Args:
            candidates (list): The list of candidates to evaluate.

        Returns:
            (list): The selected candidates.
        """
        if len(candidates) <= 1:
            return candidates

        mask = list(Trainable.keys())
        mask.append("reward")

        N = len(candidates)
        fitness_values = [c.get("reward", 0.0) for c in candidates]
        competition_scores = [0.0] * N

        for i in range(N):
            fi = fitness_values[i]

            fitter_indices = [j for j in range(N) if j != i and fitness_values[j] > fi]

            if not fitter_indices:
                competition_scores[i] = 1.0
            else:
                distances = []
                for j in fitter_indices:
                    distance = await self.distance_function(
                        out_mask_json(candidates[i], mask=mask),
                        out_mask_json(candidates[j], mask=mask),
                        embedding_model=self.embedding_model,
                        **self.kwargs,
                    )
                    distances.append((j, distance))

                distances.sort(key=lambda x: x[1])
                k = min(self.k_nearest_fitter, len(distances))
                k_nearest_distances = [d[1] for d in distances[:k]]

                competition_scores[i] = float(
                    np.sum(k_nearest_distances) / k if k > 0 else 0.0
                )
        median_score = np.median(competition_scores)
        selected_candidates = []
        for i, candidate in enumerate(candidates):
            if competition_scores[i] >= median_score:
                selected_candidates.append(candidate)
        return selected_candidates

    async def on_epoch_end(
        self,
        epoch,
        trainable_variables,
    ):
        """Called at the end of an epoch

        Args:
            epoch (int): The epoch number
            trainable_variables (list): The list of trainable variables
        """
        for trainable_variable in trainable_variables:
            candidates = trainable_variable.get("candidates")
            best_candidates = trainable_variable.get("best_candidates")
            all_candidates = candidates + best_candidates
            sorted_candidates = sorted(
                all_candidates,
                key=lambda x: x.get("reward"),
                reverse=True,
            )
            sorted_candidates = await self.competition(sorted_candidates)
            selected_candidates = sorted_candidates[: self.population_size]
            trainable_variable.update(
                {
                    "best_candidates": selected_candidates,
                }
            )
        self.increment_epochs()

    def get_config(self):
        config = {
            "k_nearest_fitter": self.k_nearest_fitter,
            "mutation_temperature": self.mutation_temperature,
            "crossover_temperature": self.crossover_temperature,
            "few_shot_learning": self.few_shot_learning,
            "nb_min_examples": self.nb_min_examples,
            "nb_max_examples": self.nb_max_examples,
            "sampling_temperature": self.sampling_temperature,
            "merging_rate": self.merging_rate,
            "population_size": self.population_size,
            "name": self.name,
            "description": self.description,
        }
        language_model_config = {
            "language_model": serialization_lib.serialize_synalinks_object(
                self.language_model,
            )
        }
        embedding_model_config = {
            "embedding_model": serialization_lib.serialize_synalinks_object(
                self.embedding_model,
            )
        }
        return {**config, **language_model_config, **embedding_model_config}

    @classmethod
    def from_config(cls, config):
        language_model = serialization_lib.deserialize_synalinks_object(
            config.pop("language_model"),
        )
        embedding_model = serialization_lib.deserialize_synalinks_object(
            config.pop("embedding_model"),
        )
        return cls(
            language_model=language_model, embedding_model=embedding_model, **config
        )



# Modified from: keras/src/ops/symbolic_arguments_test.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.backend import JsonDataModel
from synalinks.src.backend import SymbolicDataModel
from synalinks.src.ops.symbolic_arguments import SymbolicArguments


class SymbolicArgumentsTest(testing.TestCase):
    # Testing multiple args and empty kwargs
    def test_args(self):
        class Test(DataModel):
            foo1: str
            foo2: str

        a = SymbolicDataModel(data_model=Test)
        b = SymbolicDataModel(data_model=Test)
        args = SymbolicArguments(
            (
                a,
                b,
            ),
            {},
        )

        self.assertEqual(args.symbolic_data_models, [a, b])
        self.assertEqual(args._flat_arguments, [a, b])
        self.assertEqual(args._single_positional_data_model, None)

    # Testing single arg and single position data_model
    def test_args_single_arg(self):
        class Test(DataModel):
            foo1: str
            foo2: str

        a = SymbolicDataModel(data_model=Test)
        args = SymbolicArguments((a))

        self.assertEqual(args.symbolic_data_models, [a])
        self.assertEqual(args._flat_arguments, [a])
        self.assertEqual(len(args.kwargs), 0)
        self.assertEqual(isinstance(args.args[0], SymbolicDataModel), True)
        self.assertEqual(args._single_positional_data_model, a)

    # Testing kwargs
    def test_kwargs(self):
        class Test(DataModel):
            foo1: str
            foo2: str

        a = SymbolicDataModel(data_model=Test)
        b = SymbolicDataModel(data_model=Test)
        c = SymbolicDataModel(data_model=Test)
        args = SymbolicArguments(
            (
                a,
                b,
            ),
            {1: c},
        )

        self.assertEqual(args.symbolic_data_models, [a, b, c])
        self.assertEqual(args._flat_arguments, [a, b, c])
        self.assertEqual(args._single_positional_data_model, None)

    # Testing fill in function with single args only
    def test_fill_in_single_arg(self):
        class Test(DataModel):
            foo1: str
            foo2: str

        a = SymbolicDataModel(data_model=Test)

        json_data_model = JsonDataModel(data_model=Test(foo1="foo1", foo2="foo2"))

        data_model_dict = {id(a): json_data_model}
        sym_args = SymbolicArguments((a))

        # Call the method to be tested
        result, _ = sym_args.fill_in(data_model_dict)

        self.assertEqual(result, (json_data_model,))

    # Testing fill in function with multiple args
    def test_fill_in_multiple_arg(self):
        class Test(DataModel):
            foo1: str
            foo2: str

        a = SymbolicDataModel(data_model=Test)
        b = SymbolicDataModel(data_model=Test)

        json_data_model = JsonDataModel(data_model=Test(foo1="foo1", foo2="foo2"))

        data_model_dict = {id(b): json_data_model}
        sym_args = SymbolicArguments((a, b))

        # Call the method to be tested
        result, _ = sym_args.fill_in(data_model_dict)
        self.assertEqual(result, ((None, json_data_model),))

    # Testing fill in function for args and kwargs
    def test_fill_in(self):
        class Test1(DataModel):
            foo: str
            bar: int

        class Test2(DataModel):
            foo: int
            bar: str

        a = SymbolicDataModel(schema=Test1.get_schema())
        b = SymbolicDataModel(schema=Test2.get_schema())
        c = SymbolicDataModel(schema=Test2.get_schema())

        json_data_model1 = JsonDataModel(data_model=Test1(foo="foo", bar=2))
        json_data_model2 = JsonDataModel(data_model=Test2(foo=3, bar="foo"))

        dictionary = {id(a): json_data_model1, id(c): json_data_model2}

        sym_args = SymbolicArguments(
            (
                a,
                b,
            ),
            {"1": c},
        )

        (values, _) = sym_args.fill_in(dictionary)
        self.assertEqual(values, ((json_data_model1, None), {"1": json_data_model2}))



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import GenericResult
from synalinks.src.backend import JsonDataModel
from synalinks.src.backend import SymbolicDataModel
from synalinks.src.backend import any_symbolic_data_models
from synalinks.src.ops.operation import Operation
from synalinks.src.saving import serialization_lib


class UpdateKnowledge(Operation):
    def __init__(
        self,
        knowledge_base=None,
        threshold=0.8,
        name=None,
        description=None,
        **kwargs,
    ):
        super().__init__(
            name=name,
            description=description,
        )
        self.knowledge_base = knowledge_base
        self.threshold = 0.8

    async def call(self, x):
        await self.knowledge_base.update(
            x,
            threshold=self.threshold,
        )
        return JsonDataModel(
            json=x.get_json(),
            schema=x.get_schema(),
            name=self.name,
        )

    async def compute_output_spec(self, x):
        return SymbolicDataModel(
            schema=x.get_schema(),
            name=self.name,
        )

    def get_config(self):
        config = {
            "name": self.name,
            "description": self.description,
        }
        knowledge_base_config = serialization_lib.serialize_synalinks_object(
            self.knowledge_base
        )
        return {"knowledge_base": knowledge_base_config, **config}

    @classmethod
    def from_config(cls, config):
        knowledge_base = serialization_lib.deserialize_synalinks_object(
            config.pop("knowledge_base")
        )
        return cls(knowledge_base=knowledge_base, **config)


@synalinks_export(
    [
        "synalinks.ops.update_knowledge",
        "synalinks.ops.knowledge_bases.update_knowledge",
    ]
)
async def update_knowledge(
    x,
    knowledge_base=None,
    threshold=0.8,
    name=None,
    description=None,
):
    """Update the knowledge base with new data.

    This function updates the specified knowledge base with the
    data model provided in the input.

    Args:
        x (JsonDataModel | SymbolicDataModel): The data model to update the
            knowledge base with.
        knowledge_base (KnowledgeBase): The knowledge base to update. Required.
        align (bool): Wether of not to perform automatic alignment (Default to True).
        threshold (float): The alignment threshold (Default to 0.7).
        name (str): Optional name for the operation.
        description (str): Optional description for the operation.

    Returns:
        (JsonDataModel| SymbolicDataModel): The resulting json data model.

    Raises:
        ValueError: If `knowledge_base` is not provided.
    """
    if knowledge_base is None:
        raise ValueError("You should provide the `knowledge_base` argument")
    if any_symbolic_data_models(x):
        return await UpdateKnowledge(
            knowledge_base=knowledge_base,
            name=name,
            threshold=threshold,
            description=description,
        ).symbolic_call(x)
    return await UpdateKnowledge(
        knowledge_base=knowledge_base,
        name=name,
        description=description,
    )(x)


class TripletSearch(Operation):
    def __init__(
        self,
        knowledge_base=None,
        k=10,
        threshold=0.7,
        name=None,
        description=None,
        **kwargs,
    ):
        super().__init__(
            name=name,
            description=description,
        )
        self.knowledge_base = knowledge_base
        self.k = k
        self.threshold = threshold

    async def call(self, x):
        result = await self.knowledge_base.triplet_search(
            x,
            k=self.k,
            threshold=self.threshold,
        )
        return JsonDataModel(
            json={"result": result},
            schema=GenericResult.get_schema(),
            name=self.name,
        )

    async def compute_output_spec(self, x):
        return SymbolicDataModel(
            schema=GenericResult.get_schema(),
            name=self.name,
        )

    def get_config(self):
        config = {
            "k": self.k,
            "threshold": self.threshold,
            "name": self.name,
            "description": self.description,
        }
        knowledge_base_config = serialization_lib.serialize_synalinks_object(
            self.knowledge_base
        )
        return {"knowledge_base": knowledge_base_config, **config}

    @classmethod
    def from_config(cls, config):
        knowledge_base = serialization_lib.deserialize_synalinks_object(
            config.pop("knowledge_base")
        )
        return cls(knowledge_base=knowledge_base, **config)


async def triplet_search(
    x,
    knowledge_base=None,
    k=10,
    threshold=0.7,
    name=None,
    description=None,
):
    """Search for triplets in the given knowledge base.

    This function performs a hybrid search for triplets in the specified knowledge base
    using the input data as the query (see `TripletSearch` data model).

    Args:
        x (JsonDataModel | SymbolicDataModel): The query for the triplet search.
        knowledge_base (KnowledgeBase): The knowledge base to search in.
        k (int): Maximum number of results to return. Defaults to 10.
        threshold (float): Similarity threshold for filtering results. Only results with
            similarity scores above this threshold will be returned. Defaults to 0.7.
        name (str): Optional name for the operation.
        description (str): Optional description for the operation.

    Returns:
        (JsonDataModel): A data model containing the search results.

    Raises:
        ValueError: If knowledge_base is not provided.
    """
    if knowledge_base is None:
        raise ValueError("You should provide the `knowledge_base` argument")
    if any_symbolic_data_models(x):
        return await TripletSearch(
            knowledge_base=knowledge_base,
            name=name,
            description=description,
        ).symbolic_call(x)
    return await TripletSearch(
        knowledge_base=knowledge_base,
        k=k,
        threshold=threshold,
        name=name,
        description=description,
    )(x)


class SimilaritySearch(Operation):
    def __init__(
        self,
        knowledge_base=None,
        k=10,
        threshold=0.7,
        name=None,
        description=None,
    ):
        super().__init__(
            name=name,
            description=description,
        )
        self.knowledge_base = knowledge_base
        self.k = k
        self.threshold = threshold

    async def call(self, x):
        result = await self.knowledge_base.similarity_search(
            x,
            k=self.k,
            threshold=self.threshold,
        )
        return JsonDataModel(
            json={"result": result},
            schema=GenericResult.get_schema(),
            name=self.name,
        )

    async def compute_output_spec(self, x):
        return SymbolicDataModel(
            schema=GenericResult.get_schema(),
            name=self.name,
        )

    def get_config(self):
        config = {
            "k": self.k,
            "threshold": self.threshold,
            "name": self.name,
            "description": self.description,
        }
        knowledge_base_config = serialization_lib.serialize_synalinks_object(
            self.knowledge_base
        )
        return {"knowledge_base": knowledge_base_config, **config}

    @classmethod
    def from_config(cls, config):
        knowledge_base = serialization_lib.deserialize_synalinks_object(
            config.pop("knowledge_base")
        )
        return cls(knowledge_base=knowledge_base, **config)


@synalinks_export(
    [
        "synalinks.ops.similarity_search",
        "synalinks.ops.knowledge_bases.similarity_search",
    ]
)
async def similarity_search(
    x,
    knowledge_base=None,
    k=10,
    threshold=0.7,
    name=None,
    description=None,
):
    """Search for similar entities in the given knowledge base.

    This function performs a similarity search in the specified knowledge base
    using the input data as the query.

    Args:
        x (JsonDataModel | SymbolicDataModel): The query for the similarity search.
        knowledge_base (KnowledgeBase): The knowledge base to search in.
        k (int): Maximum number of results to return (Defaults to 10).
        threshold (float): Similarity threshold for filtering results. Only results with
            similarity scores above this threshold will be returned (Defaults to 0.7).
        name (str): Optional name for the operation.
        description (str): Optional description for the operation.

    Returns:
        (JsonDataModel): A data model containing the search results.

    Raises:
        ValueError: If knowledge_base is not provided.
    """
    if knowledge_base is None:
        raise ValueError("You should provide the `knowledge_base` argument")
    if any_symbolic_data_models(x):
        return await SimilaritySearch(
            knowledge_base=knowledge_base,
            k=k,
            threshold=threshold,
            name=name,
            description=description,
        ).symbolic_call(x)
    return await SimilaritySearch(
        knowledge_base=knowledge_base,
        k=k,
        threshold=threshold,
        name=name,
        description=description,
    )(x)



from synalinks.src import backend
from synalinks.src import datasets
from synalinks.src import embedding_models
from synalinks.src import initializers
from synalinks.src import language_models
from synalinks.src import metrics
from synalinks.src import modules
from synalinks.src import ops
from synalinks.src import optimizers
from synalinks.src import programs
from synalinks.src import rewards
from synalinks.src import saving
from synalinks.src import testing
from synalinks.src import trainers
from synalinks.src import tree
from synalinks.src import utils
from synalinks.src.backend import SymbolicDataModel
from synalinks.src.modules import Input
from synalinks.src.programs import Functional
from synalinks.src.programs import Program
from synalinks.src.programs import Sequential
from synalinks.src.version import __version__



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import asyncio
import logging
import time

import requests

from synalinks.src import tree
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import any_symbolic_data_models
from synalinks.src.backend import api_base
from synalinks.src.backend import api_key
from synalinks.src.hooks.hook import Hook


@synalinks_export("synalinks.hooks.Monitor")
class Monitor(Hook):
    """Monitor hook for sending module call traces to a remote endpoint in realtime.

    This hook sends trace data immediately to a specified endpoint for realtime monitoring.
    Traces are sent asynchronously using asyncio to avoid blocking module execution.

    You can enable monitoring for every modules by using `synalinks.enable_observability()`
    at the beginning of your scripts:

    Example:

    ```python
    import synalinks

    synalinks.enable_observability()
    ```

    Args:
        timeout: Request timeout in seconds (default: 5).
        headers: Optional additional headers to include in requests
    """

    def __init__(
        self,
        timeout=5,
        headers=None,
    ):
        super().__init__()
        self.endpoint = api_base()
        self.timeout = timeout
        if api_key() is not None and not headers:
            headers = {"Authorization": api_key()}
        self.headers = headers or {}
        self.call_start_times = {}
        self._pending_tasks = []
        self.logger = logging.getLogger(__name__)

    async def _post_trace(self, data: dict):
        """POST trace data to the endpoint asynchronously."""
        url = f"{self.endpoint}/trace"

        try:
            loop = asyncio.get_event_loop()
            # Run requests in executor to make it non-blocking
            response = await loop.run_in_executor(
                None,
                lambda: requests.post(
                    url,
                    json=data,
                    headers=self.headers,
                    timeout=self.timeout,
                )
            )
            response.raise_for_status()
            self.logger.debug(
                f"Trace sent successfully: {data.get('event')} for call {data.get('call_id')}"
            )
        except requests.exceptions.Timeout:
            self.logger.error(f"Timeout sending trace to {url}")
        except requests.exceptions.RequestException as e:
            self.logger.error(f"Failed to send trace to {url}: {e}")
        except Exception as e:
            self.logger.error(f"Unexpected error sending trace: {e}")

    def _send_trace_async(self, trace_data: dict):
        """Send trace asynchronously without blocking."""
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            # No event loop in current thread, create a new one
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        # Create task and store reference to prevent garbage collection
        task = loop.create_task(self._post_trace(trace_data))
        self._pending_tasks.append(task)

        # Clean up completed tasks
        self._pending_tasks = [t for t in self._pending_tasks if not t.done()]

    def _extract_data_models_info(self, data):
        """Extract data model information from inputs/outputs."""
        if not data:
            return None

        flatten_data = tree.flatten(data)

        if any_symbolic_data_models(data):
            schemas = [dm.get_schema() for dm in flatten_data if dm is not None]
            return {
                "type": "symbolic",
                "schemas": schemas,
            }
        else:
            jsons = [dm.get_json() for dm in flatten_data if dm is not None]
            return {
                "type": "data",
                "data": jsons,
            }

    def on_call_begin(
        self,
        call_id,
        inputs=None,
    ):
        """Called when a module call begins."""
        self.call_start_times[call_id] = time.time()

        trace_data = {
            "event": "call_begin",
            "call_id": call_id,
            "module_name": self.module.name,
            "module_description": self.module.description,
            "timestamp": self.call_start_times[call_id],
            "inputs": self._extract_data_models_info(inputs),
        }

        self._send_trace_async(trace_data)

    def on_call_end(
        self,
        call_id,
        outputs=None,
        exception=None,
    ):
        """Called when a module call ends."""
        end_time = time.time()
        start_time = self.call_start_times.pop(call_id, end_time)
        duration = end_time - start_time

        trace_data = {
            "event": "call_end",
            "call_id": call_id,
            "module_name": self.module.name,
            "module_description": self.module.description,
            "timestamp": end_time,
            "duration": duration,
            "outputs": self._extract_data_models_info(outputs),
            "exception": str(exception) if exception else None,
            "success": exception is None,
        }

        self._send_trace_async(trace_data)

    async def _cleanup(self):
        """Wait for pending tasks."""
        if self._pending_tasks:
            await asyncio.gather(*self._pending_tasks, return_exceptions=True)

    def __del__(self):
        """Cleanup pending traces."""
        if hasattr(self, "_pending_tasks") and self._pending_tasks:
            try:
                loop = asyncio.get_event_loop()
                if loop.is_running():
                    # If loop is running, schedule cleanup
                    loop.create_task(self._cleanup())
                else:
                    # If loop is not running, run cleanup
                    loop.run_until_complete(self._cleanup())
            except Exception:
                pass


# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from typing import List
from typing import Literal
from typing import Union

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.backend import concatenate_json
from synalinks.src.backend import factorize_json
from synalinks.src.backend import in_mask_json
from synalinks.src.backend import out_mask_json


class JsonConcatenateTest(testing.TestCase):
    def test_concatenate_identical_jsons(self):
        class Input(DataModel):
            foo: str

        class Result(DataModel):
            foo: str
            foo_1: str

        json = Input(foo="test").get_json()
        expected = Result(foo="test", foo_1="test").get_json()

        result = concatenate_json(json, json)
        self.assertEqual(result, expected)

    def test_concatenate_jsons_with_different_properties(self):
        class Input1(DataModel):
            foo: str

        class Input2(DataModel):
            bar: str

        class Result(DataModel):
            foo: str
            bar: str

        json1 = Input1(foo="test").get_json()

        json2 = Input2(bar="test").get_json()

        expected = Result(foo="test", bar="test").get_json()

        result = concatenate_json(json1, json2)
        self.assertEqual(result, expected)

    def test_concatenate_json_multiple_times(self):
        class Input(DataModel):
            foo: str

        class Result(DataModel):
            foo: str
            foo_1: str
            foo_2: str

        json = Input(foo="test").get_json()
        expected = Result(foo="test", foo_1="test", foo_2="test").get_json()

        result = concatenate_json(json, json)
        result = concatenate_json(result, json)
        self.assertEqual(result, expected)

    def test_concatenate_nested(self):
        class BarObject(DataModel):
            foo: str
            bar: str

        class Input(DataModel):
            bar: BarObject

        class Result(DataModel):
            bar: BarObject
            bar_1: BarObject

        json = Input(bar=BarObject(foo="test", bar="test")).get_json()
        expected = Result(
            bar=BarObject(foo="test", bar="test"),
            bar_1=BarObject(foo="test", bar="test"),
        ).get_json()

        result = concatenate_json(json, json)
        self.assertEqual(result, expected)

    def test_concatenate_similar_entities(self):
        class City(DataModel):
            label: Literal["City"]
            name: str

        class Cities(DataModel):
            entities: List[City]

        class Result(DataModel):
            entities: List[City]
            entities_1: List[City]

        paris = City(label="City", name="Paris")
        toulouse = City(label="City", name="Toulouse")

        json = Cities(entities=[paris, toulouse]).get_json()
        expected = Result(
            entities=[paris, toulouse],
            entities_1=[paris, toulouse],
        ).get_json()

        result = concatenate_json(json, json)
        self.assertEqual(result, expected)

    def test_concatenate_different_entities(self):
        class City(DataModel):
            label: Literal["City"]
            name: str

        class Event(DataModel):
            label: Literal["Event"]
            name: str

        class Cities(DataModel):
            entities: List[City]

        class Events(DataModel):
            entities: List[Event]

        class Result(DataModel):
            entities: List[City]
            entities_1: List[Event]

        paris = City(label="City", name="Paris")
        toulouse = City(label="City", name="Toulouse")

        event1 = Event(label="Event", name="A test event")
        event2 = Event(label="Event", name="Another test event")

        json1 = Cities(entities=[paris, toulouse]).get_json()
        json2 = Events(entities=[event1, event2]).get_json()

        expected = Result(
            entities=[paris, toulouse],
            entities_1=[event1, event2],
        ).get_json()

        result = concatenate_json(json1, json2)
        self.assertEqual(result, expected)


class JsonFactorizeTest(testing.TestCase):
    def test_factorize_json_with_identical_properties(self):
        class Input(DataModel):
            foo: str
            foo_1: str

        class Result(DataModel):
            foos: List[str]

        json = Input(foo="test", foo_1="test").get_json()
        expected = Result(foos=["test", "test"]).get_json()

        result = factorize_json(json)
        self.assertEqual(result, expected)

    def test_factorize_json_with_multiple_identical_properties(self):
        class Input(DataModel):
            foo: str
            foo_1: str
            foo_2: str

        class Result(DataModel):
            foos: List[str]

        json = Input(foo="test", foo_1="test", foo_2="test").get_json()
        expected = Result(foos=["test", "test", "test"]).get_json()

        result = factorize_json(json)
        self.assertEqual(result, expected)

    def test_factorize_json_with_different_properties(self):
        class Input(DataModel):
            foo: str
            bar: str

        class Result(DataModel):
            foo: str
            bar: str

        json = Input(foo="test", bar="test").get_json()
        expected = Result(foo="test", bar="test").get_json()

        result = factorize_json(json)
        self.assertEqual(result, expected)

    def test_factorize_json_with_mixed_properties(self):
        class Input(DataModel):
            foo: str
            foo_1: str
            bar: str
            boo: str

        class Result(DataModel):
            foos: List[str]
            bar: str
            boo: str

        json = Input(
            foo="test",
            foo_1="test",
            bar="test",
            boo="test",
        ).get_json()

        expected = Result(
            foos=["test", "test"],
            bar="test",
            boo="test",
        ).get_json()

        result = factorize_json(json)
        self.assertEqual(result, expected)

    def test_factorize_json_with_existing_array_property(self):
        class Input(DataModel):
            foos: List[str]
            foo: str

        class Result(DataModel):
            foos: List[str]

        json = Input(foos=["test"], foo="test").get_json()
        expected = Result(foos=["test", "test"]).get_json()

        result = factorize_json(json)
        self.assertEqual(result, expected)

    def test_factorize_json_with_existing_array_property_and_additional_properties(
        self,
    ):
        class Input(DataModel):
            foos: List[str]
            foo: str
            foo_1: str

        class Result(DataModel):
            foos: List[str]

        json = Input(
            foos=["test", "test"],
            foo="test",
            foo_1="test",
        ).get_json()

        expected = Result(foos=["test", "test", "test", "test"]).get_json()

        result = factorize_json(json)
        self.assertEqual(result, expected)

    def test_factorize_json_with_multiple_groups_of_properties(self):
        class Input(DataModel):
            foo: str
            foo_1: str
            bar: str
            bar_1: str

        class Result(DataModel):
            foos: List[str]
            bars: List[str]

        json = Input(
            foo="test",
            foo_1="test",
            bar="test",
            bar_1="test",
        ).get_json()

        expected = Result(
            foos=["test", "test"],
            bars=["test", "test"],
        ).get_json()

        result = factorize_json(json)
        self.assertEqual(result, expected)

    def test_factorize_nested(self):
        class BarObject(DataModel):
            foo: str
            bar: str

        class Input(DataModel):
            bar: BarObject
            bar_1: BarObject

        class Result(DataModel):
            bars: List[BarObject]

        json = Input(
            bar=BarObject(
                foo="test",
                bar="test",
            ),
            bar_1=BarObject(
                foo="test",
                bar="test",
            ),
        ).get_json()

        expected = Result(
            bars=[
                BarObject(
                    foo="test",
                    bar="test",
                ),
                BarObject(
                    foo="test",
                    bar="test",
                ),
            ]
        ).get_json()

        result = factorize_json(json)
        self.assertEqual(result, expected)

    def test_factorize_similar_entities(self):
        class City(DataModel):
            label: Literal["City"]
            name: str

        class Cities(DataModel):
            entities: List[City]

        class Input(DataModel):
            entities: List[City]
            entities_1: List[City]

        class Result(DataModel):
            entities: List[City]

        paris = City(label="City", name="Paris")
        toulouse = City(label="City", name="Toulouse")

        inputs = Input(
            entities=[paris, toulouse],
            entities_1=[paris, toulouse],
        ).get_json()
        expected = Result(
            entities=[paris, toulouse, paris, toulouse],
        ).get_json()

        result = factorize_json(inputs)
        self.assertEqual(result, expected)

    def test_factorize_different_entities(self):
        class City(DataModel):
            label: Literal["City"]
            name: str

        class Event(DataModel):
            label: Literal["Event"]
            name: str

        class Input(DataModel):
            entities: List[City]
            entities_1: List[Event]

        class Result(DataModel):
            entities: List[Union[City, Event]]

        paris = City(label="City", name="Paris")
        toulouse = City(label="City", name="Toulouse")

        event1 = Event(label="Event", name="A test event")
        event2 = Event(label="Event", name="Another test event")

        inputs = Input(
            entities=[paris, toulouse],
            entities_1=[event1, event2],
        ).get_json()
        expected = Result(
            entities=[paris, toulouse, event1, event2],
        ).get_json()

        result = factorize_json(inputs)
        self.assertEqual(result, expected)


class JsonOutMaskTest(testing.TestCase):
    def test_mask_basic(self):
        class Input(DataModel):
            foo: str
            bar: str

        class Result(DataModel):
            bar: str

        json = Input(foo="test", bar="test").get_json()
        expected = Result(bar="test").get_json()

        result = out_mask_json(json, mask=["foo"])
        self.assertEqual(result, expected)

    def test_mask_multiple_fields_with_same_base_name(self):
        class Input(DataModel):
            foo: str
            foo_1: str
            bar: str
            bar_1: str

        class Result(DataModel):
            bar: str
            bar_1: str

        json = Input(
            foo="test",
            foo_1="str",
            bar="test",
            bar_1="test",
        ).get_json()

        expected = Result(
            bar="test",
            bar_1="test",
        ).get_json()

        result = out_mask_json(json, mask=["foo"])
        self.assertEqual(result, expected)

    def test_mask_nested(self):
        class BarObject(DataModel):
            foo: str
            bar: str

        class Input(DataModel):
            bar: BarObject
            bar_1: BarObject

        json = Input(
            bar=BarObject(
                foo="test",
                bar="test",
            ),
            bar_1=BarObject(
                foo="test",
                bar="test",
            ),
        ).get_json()

        class BarObject(DataModel):
            bar: str

        class Result(DataModel):
            bar: BarObject
            bar_1: BarObject

        expected = Result(
            bar=BarObject(
                bar="test",
            ),
            bar_1=BarObject(
                bar="test",
            ),
        ).get_json()

        result = out_mask_json(json, mask=["foo"])
        self.assertEqual(result, expected)

    def test_mask_deeply_nested(self):
        class BooObject(DataModel):
            foo: str
            boo: str

        class BarObject(DataModel):
            boo: BooObject

        class Input(DataModel):
            foo: str
            bar: BarObject

        json = Input(
            foo="test",
            bar=BarObject(
                boo=BooObject(
                    foo="test",
                    boo="test",
                )
            ),
        ).get_json()

        class BooObject(DataModel):
            boo: str

        class BarObject(DataModel):
            boo: BooObject

        class Result(DataModel):
            bar: BarObject

        expected = Result(
            bar=BarObject(
                boo=BooObject(
                    boo="test",
                )
            )
        ).get_json()

        result = out_mask_json(json, mask=["foo"])
        self.assertEqual(result, expected)

    def test_mask_in_array(self):
        class BooObject(DataModel):
            foo: str
            boo: str

        class Input(DataModel):
            boos: List[BooObject]

        json = Input(
            boos=[
                BooObject(
                    foo="test",
                    boo="test",
                ),
                BooObject(
                    foo="test",
                    boo="test",
                ),
            ]
        ).get_json()

        class BooObject(DataModel):
            boo: str

        class Result(DataModel):
            boos: List[BooObject]

        expected = Result(
            boos=[
                BooObject(
                    boo="test",
                ),
                BooObject(
                    boo="test",
                ),
            ]
        ).get_json()

        result = out_mask_json(json, mask=["foo"])
        self.assertEqual(result, expected)

    def test_mask_empty_json(self):
        class Input(DataModel):
            pass

        class Result(DataModel):
            pass

        json = Input().get_json()
        expected = Result().get_json()

        result = out_mask_json(json, mask=["foo"])
        self.assertEqual(result, expected)

    def test_mask_empty_mask_list(self):
        class Input(DataModel):
            foo: str
            bar: str

        class Result(DataModel):
            foo: str
            bar: str

        json = Input(foo="test", bar="test").get_json()
        expected = Result(foo="test", bar="test").get_json()

        result = out_mask_json(json, mask=[])
        self.assertEqual(result, expected)

    def test_mask_non_recursive(self):
        class BooObject(DataModel):
            foo: str
            boo: str

        class Input(DataModel):
            foo: str
            boo: BooObject

        json = Input(
            foo="test",
            boo=BooObject(foo="test", boo="test"),
        ).get_json()

        class Result(DataModel):
            boo: BooObject

        expected = Result(
            boo=BooObject(
                foo="test",
                boo="test",
            )
        ).get_json()

        result = out_mask_json(json, mask=["foo"], recursive=False)
        self.assertEqual(result, expected)


class JsonInMaskTest(testing.TestCase):
    def test_mask_basic(self):
        json = {
            "foo": "test",
            "bar": "test",
        }

        expected = {
            "foo": "test",
        }

        result = in_mask_json(json, mask=["foo"])
        self.assertEqual(result, expected)

    def test_mask_keep_all(self):
        json = {
            "foo": "test",
            "foos": "test",
            "bar": "test",
        }

        expected = {
            "foo": "test",
            "foos": "test",
            "bar": "test",
        }

        result = in_mask_json(json, mask=["foos", "bar"])
        self.assertEqual(result, expected)

    def test_mask_multiple_fields_with_same_base_name(self):
        json = {"foo": "test", "foo_1": "test", "bar": "test", "bar_1": "test"}

        expected = {"foo": "test", "foo_1": "test"}

        result = in_mask_json(json, mask=["foo"])
        self.assertEqual(result, expected)

    def test_mask_nested(self):
        json = {
            "foo": "test",
            "foo_1": "test",
            "bar": {"foo": "test", "bar": "test"},
            "bar_1": "test",
        }

        expected = {"foo": "test", "foo_1": "test"}

        result = in_mask_json(json, mask=["foo"])
        self.assertEqual(result, expected)

    def test_mask_deeply_nested(self):
        json = {"foo": "test", "bar": {"bar": {"foo": "test", "qux": "test"}}}

        expected = {"foo": "test", "bar": {"bar": {"foo": "test"}}}

        result = in_mask_json(json, mask=["foo", "bar"])
        self.assertEqual(result, expected)

    def test_mask_in_array(self):
        json = {
            "items": [
                {"foo": "test", "bar": "test"},
                {"foo_1": "test", "bar_1": "test"},
            ]
        }

        expected = {"items": [{"foo": "test"}, {"foo_1": "test"}]}

        result = in_mask_json(json, mask=["foo"])
        self.assertEqual(result, expected)

    def test_mask_empty_json(self):
        json = {}
        expected = {}

        result = in_mask_json(json, mask=["foo"])
        self.assertEqual(result, expected)

    def test_mask_empty_mask_list(self):
        json = {"foo": "test", "bar": "test"}

        expected = {}

        result = in_mask_json(json, mask=[])
        self.assertEqual(result, expected)

    def test_mask_non_recursive(self):
        json = {"foo": "test", "bar": {"foo": "test", "bar": "test"}}

        expected = {"foo": "test"}

        result = in_mask_json(json, mask=["foo"], recursive=False)
        self.assertEqual(result, expected)



# Modified from: keras/src/layers/core/input_layer.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import backend
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import is_schema_equal
from synalinks.src.backend import is_symbolic_data_model
from synalinks.src.backend import standardize_schema
from synalinks.src.modules.module import Module
from synalinks.src.ops.node import Node


@synalinks_export("synalinks.modules.InputModule")
class InputModule(Module):
    def __init__(
        self,
        schema=None,
        input_data_model=None,
        optional=False,
        name=None,
        **kwargs,
    ):
        super().__init__(
            name=name, description="Defines the input data model for a program."
        )
        if input_data_model is not None:
            if not is_symbolic_data_model(input_data_model):
                raise ValueError(
                    "Argument `input_data_model` must be a SymbolicDataModel. "
                    f"Received invalid type: input_data_model={input_data_model} "
                    f"(of type {type(input_data_model)})"
                )
            if schema is not None:
                if not is_schema_equal(schema, input_data_model.get_schema()):
                    raise ValueError(
                        "When providing the `input_data_model` argument, you "
                        "cannot provide an incompatible `schema` argument."
                    )
            schema = input_data_model.get_schema()
        else:
            if schema is None:
                raise ValueError("You must pass a `schema` argument.")

        self._schema = standardize_schema(schema)
        if input_data_model is None:
            input_data_model = backend.SymbolicDataModel(
                schema=self._schema,
                name=name,
            )
        self._input_data_model = input_data_model
        Node(
            operation=self,
            call_args={},
            call_kwargs={},
            outputs=input_data_model,
        )
        self.built = True
        self.optional = optional

    @property
    def input_schema(self):
        return self._schema

    async def call(self):
        return

    def get_schema(self):
        return self._schema

    def get_config(self):
        return {
            "schema": self._schema,
            "optional": self.optional,
            "name": self.name,
            "description": self.description,
        }

    @classmethod
    def from_config(cls, config):
        return cls(**config)


@synalinks_export(["synalinks.modules.Input", "synalinks.Input"])
def Input(
    schema=None,
    data_model=None,
    optional=False,
    name=None,
):
    """Used to instantiate a `SymbolicDataModel`.

    A `SymbolicDataModel` is a symbolic data model-like object, which we augment with
    certain attributes that allow us to build a Synalinks `Program` just by knowing the
    inputs and outputs of the program (similar to Keras symbolic tensor).

    Example:

    ```python
    import synalinks

    class Query(synalinks.DataModel):
        query: str

    inputs = synalinks.Input(data_model=Query)

    # You can also create it using a JSON schema like this:

    inputs = synalinks.Input(schema=Query.get_schema())

    # Or using a symbolic datamodel:

    inputs = synalinks.Input(data_model=Query.to_symbolic_data_model())
    ```

    Args:
        schema (dict): A Json schema of the data_model.
            If not provided uses the `data_model` argument.
        data_model (DataModel): Optional existing data model to wrap into
            the `Input` layer. If set, the module will use this data_model rather
            than creating a new placeholder data model.
        optional (bool): Whether the input is optional or not.
            An optional input can accept `None` values.
        name (string): Optional name string for the module.
            Should be unique in a program (do not reuse the same name twice).
            It will be autogenerated if it isn't provided.

    Returns:
        (SymbolicDataModel): The symbolic data model corresponding to
            the given data model/schema.
    """
    module = InputModule(
        schema=schema,
        input_data_model=data_model.to_symbolic_data_model() if data_model else None,
        optional=optional,
        name=name,
    )
    return module.output



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import asyncio
import copy
import json
import warnings

import litellm

from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import ChatRole
from synalinks.src.saving import serialization_lib
from synalinks.src.saving.synalinks_saveable import SynalinksSaveable
from synalinks.src.utils.nlp_utils import shorten_text

litellm.drop_params = True
litellm.disable_aiohttp_transport = True

@synalinks_export(
    [
        "synalinks.LanguageModel",
        "synalinks.language_models.LanguageModel",
    ]
)
class LanguageModel(SynalinksSaveable):
    """A language model API wrapper.

    A language model is a type of AI model designed to generate, and interpret human
    language. It is trained on large amounts of text data to learn patterns and
    structures in language. Language models can perform various tasks such as text
    generation, translation, summarization, and answering questions.

    We support providers that implement *constrained structured output*
    like OpenAI, Azure, Ollama or Mistral. In addition we support providers that otherwise
    allow to constrain the use of a specific tool like Groq or Anthropic.

    For the complete list of models, please refer to the providers documentation.

    **Using OpenAI models**

    ```python
    import synalinks
    import os

    os.environ["OPENAI_API_KEY"] = "your-api-key"

    language_model = synalinks.LanguageModel(
        model="openai/gpt-4o-mini",
    )
    ```

    **Using Groq models**

    ```python
    import synalinks
    import os

    os.environ["GROQ_API_KEY"] = "your-api-key"

    language_model = synalinks.LanguageModel(
        model="groq/llama3-8b-8192",
    )
    ```

    **Using Anthropic models**

    ```python
    import synalinks
    import os

    os.environ["ANTHROPIC_API_KEY"] = "your-api-key"

    language_model = synalinks.LanguageModel(
        model="anthropic/claude-3-sonnet-20240229",
    )
    ```

    **Using Mistral models**

    ```python
    import synalinks
    import os

    os.environ["MISTRAL_API_KEY"] = "your-api-key"

    language_model = synalinks.LanguageModel(
        model="mistral/codestral-latest",
    )
    ```

    **Using Ollama models**

    ```python
    import synalinks
    import os

    language_model = synalinks.LanguageModel(
        model="ollama/mistral",
    )
    ```

    **Using Azure OpenAI models**

    ```python
    import synalinks
    import os

    os.environ["AZURE_API_KEY"] = "your-api-key"
    os.environ["AZURE_API_BASE"] = "your-api-key"
    os.environ["AZURE_API_VERSION"] = "your-api-key"

    language_model = synalinks.LanguageModel(
        model="azure/<your_deployment_name>",
    )
    ```

    **Using Google Gemini models**

    ```python
    import synalinks
    import os

    os.environ["GEMINI_API_KEY"] = "your-api-key"

    language_model = synalinks.LanguageModel(
        model="gemini/gemini-1.5-pro",
    )
    ```

    **Using XAI models**

    ```python
    import synalinks
    import os

    os.environ["XAI_API_KEY"] = "your-api-key"

    language_model = synalinks.LanguageModel(
        model="xai/grok-code-fast-1",
    )
    ```

    To cascade models in case there is anything wrong with
    the model provider (hence making your pipelines more robust).
    Use the `fallback` argument like in this example:

    ```python
    import synalinks
    import os

    os.environ["OPENAI_API_KEY"] = "your-api-key"
    os.environ["ANTHROPIC_API_KEY"] = "your-api-key"

    language_model = synalinks.LanguageModel(
        model="anthropic/claude-3-sonnet-20240229",
        fallback=synalinks.LanguageModel(
            model="openai/gpt-4o-mini",
        )
    )
    ```

    **Note**: Obviously, use an `.env` file and `.gitignore` to avoid
    putting your API keys in the code or a config file that can lead to
    leackage when pushing it into repositories.

    Args:
        model (str): The model to use.
        api_base (str): Optional. The endpoint to use.
        timeout (int): Optional. The timeout value in seconds (Default to 600).
        retry (int): Optional. The number of retry (default to 5).
        fallback (LanguageModel): Optional. The language model to fallback
            if anything is wrong.
        caching (bool): Optional. Enable caching of LM calls (Default to False).
    """

    def __init__(
        self,
        model=None,
        api_base=None,
        timeout=600,
        retry=5,
        fallback=None,
        caching=False,
    ):
        if model is None:
            raise ValueError("You need to set the `model` argument for any LanguageModel")
        model_provider = model.split("/")[0]
        if model_provider == "ollama":
            # Switch from `ollama` to `ollama_chat`
            # because it have better performance due to the chat prompts
            model = model.replace("ollama", "ollama_chat")
        self.model = model
        self.fallback = fallback
        if self.model.startswith("ollama") and not api_base:
            self.api_base = "http://localhost:11434"
        else:
            self.api_base = api_base
        self.timeout = timeout
        self.retry = retry
        self.caching = caching

    async def __call__(self, messages, schema=None, streaming=False, **kwargs):
        """
        Call method to generate a response using the language model.

        Args:
            messages (dict): A formatted dict of chat messages.
            schema (dict): The target JSON schema for structed output (optional).
                If None, output a ChatMessage-like answer.
            streaming (bool): Enable streaming (optional). Default to False.
                Can be enabled only if schema is None.
            **kwargs (keyword arguments): The additional keywords arguments
                forwarded to the LM call.
        Returns:
            (dict): The generated structured response.
        """
        formatted_messages = messages.get_json().get("messages", [])
        json_instance = {}
        input_kwargs = copy.deepcopy(kwargs)
        schema = copy.deepcopy(schema)
        if schema:
            if self.model.startswith("groq"):
                # Use a tool created on the fly for groq
                kwargs.update(
                    {
                        "tools": [
                            {
                                "function": {
                                    "name": "structured_output",
                                    "description": "Generate a valid JSON output",
                                    "parameters": schema.get("properties"),
                                },
                                "type": "function",
                            }
                        ],
                        "tool_choice": {
                            "type": "function",
                            "function": {"name": "structured_output"},
                        },
                    }
                )
            elif self.model.startswith("anthropic"):
                # Use a tool created on the fly for anthropic
                kwargs.update(
                    {
                        "tools": [
                            {
                                "name": "structured_output",
                                "description": "Generate a valid JSON output",
                                "input_schema": {
                                    "type": "object",
                                    "properties": schema.get("properties"),
                                    "required": schema.get("required"),
                                },
                            }
                        ],
                        "tool_choice": {
                            "type": "tool",
                            "name": "structured_output",
                        },
                    }
                )
            elif self.model.startswith("ollama") or self.model.startswith("mistral"):
                # Use constrained structured output for ollama/mistral
                kwargs.update(
                    {
                        "response_format": {
                            "type": "json_schema",
                            "json_schema": {"schema": schema},
                            "strict": True,
                        },
                    }
                )
            elif self.model.startswith("openai") or self.model.startswith("azure"):
                # Use constrained structured output for openai
                # OpenAI require the field  "additionalProperties"
                # Also OpenAI disallow the field "description" in $ref
                if "properties" in schema:
                    for prop_key, prop_value in schema["properties"].items():
                        if "$ref" in prop_value and "description" in prop_value:
                            del prop_value["description"]
                kwargs.update(
                    {
                        "response_format": {
                            "type": "json_schema",
                            "json_schema": {
                                "name": "structured_output",
                                "strict": True,
                                "schema": schema,
                            },
                        }
                    }
                )
            elif self.model.startswith("gemini"):
                kwargs.update(
                    {
                        "response_format": {
                            "type": "json_schema",
                            "json_schema": {
                                "schema": schema,
                            },
                            "strict": True,
                        }
                    }
                )
            elif self.model.startswith("xai"):
                kwargs.update(
                    {
                        "response_format": {
                            "type": "json_schema",
                            "json_schema": {
                                "schema": schema,
                            },
                            "strict": True,
                        }
                    }
                )
            else:
                provider = self.model.split("/")[0]
                raise ValueError(
                    f"LM provider '{provider}' not supported yet, please ensure that"
                    " they support constrained structured output and fill an issue."
                )

        if self.api_base:
            kwargs.update(
                {
                    "api_base": self.api_base,
                }
            )
        if streaming and schema:
            streaming = False
        if streaming:
            kwargs.update({"stream": True})
        for i in range(self.retry):
            try:
                response_str = ""
                response = await litellm.acompletion(
                    model=self.model,
                    messages=formatted_messages,
                    timeout=self.timeout,
                    caching=self.caching,
                    **kwargs,
                )
                if streaming:
                    return StreamingIterator(response)
                if (
                    self.model.startswith("groq") or self.model.startswith("anthropic")
                ) and schema:
                    response_str = response["choices"][0]["message"]["tool_calls"][0][
                        "function"
                    ]["arguments"]
                else:
                    response_str = response["choices"][0]["message"]["content"].strip()
                if schema:
                    json_instance = json.loads(response_str)
                else:
                    json_instance = {
                        "role": ChatRole.ASSISTANT,
                        "content": response_str,
                        "tool_call_id": None,
                        "tool_calls": [],
                    }
                return json_instance
            except Exception as e:
                warnings.warn(
                    f"Error occured while trying to call {self}: "
                    + str(e)
                    + f"\nReceived response={shorten_text(response_str)}"
                )
            await asyncio.sleep(1)
        if self.fallback:
            return self.fallback(
                messages,
                schema=schema,
                streaming=streaming,
                **input_kwargs,
            )
        else:
            return None

    def _obj_type(self):
        return "LanguageModel"

    def get_config(self):
        config = {
            "model": self.model,
            "api_base": self.api_base,
            "timeout": self.timeout,
            "retry": self.retry,
            "caching": self.caching,
        }
        if self.fallback:
            fallback_config = {
                "fallback": serialization_lib.serialize_synalinks_object(
                    self.fallback,
                )
            }
            return {**fallback_config, **config}
        else:
            return config

    @classmethod
    def from_config(cls, config):
        if "fallback" in config:
            fallback = serialization_lib.deserialize_synalinks_object(
                config.pop("fallback")
            )
            return cls(fallback=fallback, **config)
        else:
            return cls(**config)

    def __repr__(self):
        api_base = f" api_base={self.api_base}" if self.api_base else ""
        return f"<LanguageModel model={self.model}{api_base}>"


class StreamingIterator:
    def __init__(self, iterator):
        self._iterator = iterator

    def __iter__(self):
        return self

    def __next__(self):
        content = self._iterator.__next__()["choices"][0]["delta"]["content"]
        if content:
            return {"role": ChatRole.ASSISTANT, "content": content}
        else:
            raise StopIteration



from synalinks.src.callbacks.callback import Callback
from synalinks.src.callbacks.callback_list import CallbackList
from synalinks.src.callbacks.history import History
from synalinks.src.callbacks.progbar_logger import ProgbarLogger



# Modified from: keras/src/losses/losses.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src.api_export import synalinks_export
from synalinks.src.rewards.reward import Reward
from synalinks.src.saving import serialization_lib


@synalinks_export("synalinks.rewards.RewardFunctionWrapper")
class RewardFunctionWrapper(Reward):
    """Wrap a stateless function into a `Reward`.

    You can use this to quickly build a reward from a function. The function needs
    to have the signature `fn(y_true, y_pred)`.

    Example:

    ```python

    def my_reward(y_true, y_pred):
        # ...
        return reward

    program.compile(
        reward=synalinks.rewards.RewardFunctionWrapper,
        optimizer=synalinks.optimizers.RandomFewShot()
    )
    ```

    Args:
        fn (callable): The reward function to wrap, with signature
            `fn(y_true, y_pred, **kwargs)`.
        name (str): Optional. string name of the reward instance.
        in_mask (list): Optional. list of keys to keep to compute the reward.
        out_mask (list): Optional. list of keys to remove to compute the reward.
        **kwargs (keyword arguments): Keyword arguments to pass on to `fn`.
    """

    def __init__(
        self,
        fn,
        reduction="mean",
        name=None,
        in_mask=None,
        out_mask=None,
        **kwargs,
    ):
        super().__init__(
            name=name,
            reduction=reduction,
            in_mask=in_mask,
            out_mask=out_mask,
        )
        self.fn = fn
        self._fn_kwargs = kwargs

    async def call(self, y_true, y_pred):
        return await self.fn(y_true, y_pred, **self._fn_kwargs)

    def get_config(self):
        config = super().get_config()
        config.update({"fn": serialization_lib.serialize_synalinks_object(self.fn)})
        config.update(serialization_lib.serialize_synalinks_object(self._fn_kwargs))
        return config

    @classmethod
    def from_config(cls, config):
        if "fn" in config:
            config = serialization_lib.deserialize_synalinks_object(config)
        return cls(**config)

    def __repr__(self):
        return f"<RewardFunctionWrapper({self.fn}, kwargs={self._fn_kwargs})>"


@synalinks_export(
    [
        "synalinks.ProgramAsJudge",
        "synalinks.rewards.ProgramAsJudge",
    ]
)
class ProgramAsJudge(Reward):
    """Wrap a `Program` into a `Reward`.

    You can use this to create advanced reward functions that use a Synalinks `Program`.
    The program should have two inputs and one output.

    **Note:** The output data model/schema should have a field named `reward`.

    Example:

    ```python
    # ... your program declaration

    program = synalinks.Program(
        inputs=x0,
        outputs=xn,
    )

    program.compile(
        reward=synalinks.rewards.ProgramAsJudge(program=program)
        optimizer=synalinks.optimizers.RandomFewShot(),
    )
    ```

    Args:
        program (Program): The reward program to wrap.
        name (str): Optional. string name of the reward instance.
        in_mask (list): Optional. list of keys to keep to compute the reward.
        out_mask (list): Optional. list of keys to remove to compute the reward.
    """

    def __init__(
        self,
        program,
        reduction="mean",
        name=None,
        in_mask=None,
        out_mask=None,
    ):
        super().__init__(
            name=name,
            reduction=reduction,
            in_mask=in_mask,
            out_mask=out_mask,
        )
        self.program = program

    async def call(self, y_true, y_pred):
        result = await self.program([y_true, y_pred])
        return float(result.get("reward", 0.0))

    def get_config(self):
        config = super().get_config()
        config.update({"program": self.program})
        return config

    def __repr__(self):
        return f"<ProgramAsJudge({self.program})>"



import os
from contextlib import asynccontextmanager
from datetime import timedelta
from pathlib import Path
from typing import Any
from typing import AsyncIterator
from typing import Literal
from typing import Protocol
from typing import TypedDict

import httpx
from mcp import ClientSession
from mcp import StdioServerParameters
from mcp.client.sse import sse_client
from mcp.client.stdio import stdio_client
from mcp.client.streamable_http import streamablehttp_client

EncodingErrorHandler = Literal["strict", "ignore", "replace"]

DEFAULT_ENCODING = "utf-8"
DEFAULT_ENCODING_ERROR_HANDLER: EncodingErrorHandler = "strict"

DEFAULT_HTTP_TIMEOUT = 5
DEFAULT_SSE_READ_TIMEOUT = 60 * 5

DEFAULT_STREAMABLE_HTTP_TIMEOUT = timedelta(seconds=30)
DEFAULT_STREAMABLE_HTTP_SSE_READ_TIMEOUT = timedelta(seconds=60 * 5)


class McpHttpClientFactory(Protocol):
    def __call__(
        self,
        headers: dict[str, str] | None = None,
        timeout: httpx.Timeout | None = None,
        auth: httpx.Auth | None = None,
    ) -> httpx.AsyncClient: ...


class StdioConnection(TypedDict):
    transport: Literal["stdio"]

    command: str
    """The executable to run to start the server."""

    args: list[str]
    """Command line arguments to pass to the executable."""

    env: dict[str, str] | None
    """The environment to use when spawning the process."""

    cwd: str | Path | None
    """The working directory to use when spawning the process."""

    encoding: str
    """The text encoding used when sending/receiving messages to the server."""

    encoding_error_handler: EncodingErrorHandler
    """
    The text encoding error handler.

    See https://docs.python.org/3/library/codecs.html#codec-base-classes for
    explanations of possible values.
    """

    session_kwargs: dict[str, Any] | None
    """Additional keyword arguments to pass to the ClientSession."""


class SSEConnection(TypedDict):
    transport: Literal["sse"]

    url: str
    """The URL of the SSE endpoint to connect to."""

    headers: dict[str, Any] | None
    """HTTP headers to send to the SSE endpoint."""

    timeout: float
    """HTTP timeout."""

    sse_read_timeout: float
    """SSE read timeout."""

    session_kwargs: dict[str, Any] | None
    """Additional keyword arguments to pass to the ClientSession."""

    httpx_client_factory: McpHttpClientFactory | None
    """Custom factory for httpx.AsyncClient (optional)."""


class StreamableHttpConnection(TypedDict):
    transport: Literal["streamable_http"]

    url: str
    """The URL of the endpoint to connect to."""

    headers: dict[str, Any] | None
    """HTTP headers to send to the endpoint."""

    timeout: timedelta
    """HTTP timeout."""

    sse_read_timeout: timedelta
    """How long (in seconds) the client will wait for a new event before disconnecting.
    All other HTTP operations are controlled by `timeout`."""

    terminate_on_close: bool
    """Whether to terminate the session on close."""

    session_kwargs: dict[str, Any] | None
    """Additional keyword arguments to pass to the ClientSession."""

    httpx_client_factory: McpHttpClientFactory | None
    """Custom factory for httpx.AsyncClient (optional)."""


class WebsocketConnection(TypedDict):
    transport: Literal["websocket"]

    url: str
    """The URL of the Websocket endpoint to connect to."""

    session_kwargs: dict[str, Any] | None
    """Additional keyword arguments to pass to the ClientSession"""


Connection = (
    StdioConnection | SSEConnection | StreamableHttpConnection | WebsocketConnection
)


@asynccontextmanager
async def _create_stdio_session(
    *,
    command: str,
    args: list[str],
    env: dict[str, str] | None = None,
    cwd: str | Path | None = None,
    encoding: str = DEFAULT_ENCODING,
    encoding_error_handler: Literal[
        "strict", "ignore", "replace"
    ] = DEFAULT_ENCODING_ERROR_HANDLER,
    session_kwargs: dict[str, Any] | None = None,
) -> AsyncIterator[ClientSession]:
    """Create a new session to an MCP server using stdio.

    Args:
        command: Command to execute
        args: Arguments for the command
        env: Environment variables for the command
        cwd: Working directory for the command
        encoding: Character encoding
        encoding_error_handler: How to handle encoding errors
        session_kwargs: Additional keyword arguments to pass to the ClientSession
    """
    # NOTE: execution commands (e.g., `uvx` / `npx`) require PATH envvar to be set.
    # To address this, we automatically inject existing PATH envvar into the `env` value,
    # if it's not already set.
    env = env or {}
    if "PATH" not in env:
        env["PATH"] = os.environ.get("PATH", "")

    server_params = StdioServerParameters(
        command=command,
        args=args,
        env=env,
        cwd=cwd,
        encoding=encoding,
        encoding_error_handler=encoding_error_handler,
    )

    # Create and store the connection
    async with stdio_client(server_params) as (read, write):
        async with ClientSession(read, write, **(session_kwargs or {})) as session:
            yield session


@asynccontextmanager
async def _create_sse_session(
    *,
    url: str,
    headers: dict[str, Any] | None = None,
    timeout: float = DEFAULT_HTTP_TIMEOUT,
    sse_read_timeout: float = DEFAULT_SSE_READ_TIMEOUT,
    session_kwargs: dict[str, Any] | None = None,
    httpx_client_factory: McpHttpClientFactory | None = None,
) -> AsyncIterator[ClientSession]:
    """Create a new session to an MCP server using SSE.

    Args:
        url: (str) URL of the SSE server
        headers (dict): HTTP headers to send to the SSE endpoint
        timeout: HTTP timeout
        sse_read_timeout: SSE read timeout
        session_kwargs: Additional keyword arguments to pass to the ClientSession
        httpx_client_factory: Custom factory for httpx.AsyncClient (optional)
    """
    # Create and store the connection
    kwargs = {}
    if httpx_client_factory is not None:
        kwargs["httpx_client_factory"] = httpx_client_factory

    async with sse_client(url, headers, timeout, sse_read_timeout, **kwargs) as (
        read,
        write,
    ):
        async with ClientSession(read, write, **(session_kwargs or {})) as session:
            yield session


@asynccontextmanager
async def _create_streamable_http_session(
    *,
    url: str,
    headers: dict[str, Any] | None = None,
    timeout: timedelta = DEFAULT_STREAMABLE_HTTP_TIMEOUT,
    sse_read_timeout: timedelta = DEFAULT_STREAMABLE_HTTP_SSE_READ_TIMEOUT,
    terminate_on_close: bool = True,
    session_kwargs: dict[str, Any] | None = None,
    httpx_client_factory: McpHttpClientFactory | None = None,
) -> AsyncIterator[ClientSession]:
    """Create a new session to an MCP server using Streamable HTTP.

    Args:
        url (str): URL of the endpoint to connect to
        headers (dict): HTTP headers to send to the endpoint
        timeout: HTTP timeout
        sse_read_timeout: How long (in seconds) the client will wait for a new event
            before disconnecting
        terminate_on_close: Whether to terminate the session on close
        session_kwargs: Additional keyword arguments to pass to the ClientSession
        httpx_client_factory: Custom factory for httpx.AsyncClient (optional)
    """
    # Create and store the connection
    kwargs = {}
    if httpx_client_factory is not None:
        kwargs["httpx_client_factory"] = httpx_client_factory

    async with streamablehttp_client(
        url, headers, timeout, sse_read_timeout, terminate_on_close, **kwargs
    ) as (read, write, _):
        async with ClientSession(read, write, **(session_kwargs or {})) as session:
            yield session


@asynccontextmanager
async def _create_websocket_session(
    *,
    url: str,
    session_kwargs: dict[str, Any] | None = None,
) -> AsyncIterator[ClientSession]:
    """Create a new session to an MCP server using Websockets.

    Args:
        url: URL of the Websocket endpoint
        session_kwargs: Additional keyword arguments to pass to the ClientSession

    Raises:
        ImportError: If websockets package is not installed
    """
    try:
        from mcp.client.websocket import websocket_client
    except ImportError:
        raise ImportError(
            "Could not import websocket_client. ",
            "To use Websocket connections, please install the required dependency with: ",
            "'pip install mcp[ws]' or 'pip install websockets'",
        ) from None

    async with websocket_client(url) as (read, write):
        async with ClientSession(read, write, **(session_kwargs or {})) as session:
            yield session


@asynccontextmanager
async def create_session(
    connection: Connection,
) -> AsyncIterator[ClientSession]:
    """Create a new session to an MCP server.

    Args:
        connection: Connection config to use to connect to the server

    Raises:
        ValueError: If transport is not recognized
        ValueError: If required parameters for the specified transport are missing

    Yields:
        A ClientSession
    """

    if "transport" not in connection:
        raise ValueError(
            "Configuration error: Missing 'transport' key in server configuration. "
            "Each server must include 'transport' with one of: "
            "'stdio', 'sse', 'websocket', 'streamable_http'. "
        )

    transport = connection["transport"]
    if transport == "sse":
        if "url" not in connection:
            raise ValueError("'url' parameter is required for SSE connection")
        async with _create_sse_session(
            url=connection["url"],
            headers=connection.get("headers"),
            timeout=connection.get("timeout", DEFAULT_HTTP_TIMEOUT),
            sse_read_timeout=connection.get("sse_read_timeout", DEFAULT_SSE_READ_TIMEOUT),
            session_kwargs=connection.get("session_kwargs"),
            httpx_client_factory=connection.get("httpx_client_factory"),
        ) as session:
            yield session
    elif transport == "streamable_http":
        if "url" not in connection:
            raise ValueError("'url' parameter is required for Streamable HTTP connection")
        async with _create_streamable_http_session(
            url=connection["url"],
            headers=connection.get("headers"),
            timeout=connection.get("timeout", DEFAULT_STREAMABLE_HTTP_TIMEOUT),
            sse_read_timeout=connection.get(
                "sse_read_timeout", DEFAULT_STREAMABLE_HTTP_SSE_READ_TIMEOUT
            ),
            session_kwargs=connection.get("session_kwargs"),
            httpx_client_factory=connection.get("httpx_client_factory"),
        ) as session:
            yield session
    elif transport == "stdio":
        if "command" not in connection:
            raise ValueError("'command' parameter is required for stdio connection")
        if "args" not in connection:
            raise ValueError("'args' parameter is required for stdio connection")
        async with _create_stdio_session(
            command=connection["command"],
            args=connection["args"],
            env=connection.get("env"),
            cwd=connection.get("cwd"),
            encoding=connection.get("encoding", DEFAULT_ENCODING),
            encoding_error_handler=connection.get(
                "encoding_error_handler", DEFAULT_ENCODING_ERROR_HANDLER
            ),
            session_kwargs=connection.get("session_kwargs"),
        ) as session:
            yield session
    elif transport == "websocket":
        if "url" not in connection:
            raise ValueError("'url' parameter is required for Websocket connection")
        async with _create_websocket_session(
            url=connection["url"],
            session_kwargs=connection.get("session_kwargs"),
        ) as session:
            yield session
    else:
        raise ValueError(
            f"Unsupported transport: {transport}. "
            f"Must be one of: 'stdio', 'sse', 'websocket', 'streamable_http'"
        )



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.saving.object_registration import (
    CustomObjectScope as CustomObjectScope,
)
from synalinks.src.saving.object_registration import (
    CustomObjectScope as custom_object_scope,
)
from synalinks.src.saving.object_registration import (
    get_custom_objects as get_custom_objects,
)
from synalinks.src.saving.object_registration import (
    get_registered_name as get_registered_name,
)
from synalinks.src.saving.object_registration import (
    get_registered_object as get_registered_object,
)
from synalinks.src.saving.object_registration import (
    register_synalinks_serializable as register_synalinks_serializable,
)
from synalinks.src.saving.serialization_lib import (
    deserialize_synalinks_object as deserialize_synalinks_object,
)
from synalinks.src.saving.serialization_lib import (
    serialize_synalinks_object as serialize_synalinks_object,
)



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.initializers import deserialize as deserialize
from synalinks.src.initializers import get as get
from synalinks.src.initializers import serialize as serialize
from synalinks.src.initializers.empty_initializer import Empty as Empty
from synalinks.src.initializers.empty_initializer import Empty as empty
from synalinks.src.initializers.initializer import Initializer as Initializer



# Modified from: keras/src/metrics/f_score_metrics.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from typing import List
from typing import Optional

from synalinks.src import backend
from synalinks.src import ops
from synalinks.src import tree
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import DataModel
from synalinks.src.backend.common import numpy as np
from synalinks.src.metrics.metric import Metric
from synalinks.src.utils import nlp_utils


class FBetaState(DataModel):
    true_positives: Optional[List[float]] = None
    false_positives: Optional[List[float]] = None
    false_negatives: Optional[List[float]] = None
    intermediate_weights: Optional[List[float]] = None


@synalinks_export("synalinks.metrics.FBetaScore")
class FBetaScore(Metric):
    """Computes F-Beta score.

    Formula:

    ```python
    b2 = beta ** 2
    f_beta_score = (1 + b2) * (precision * recall) / (precision * b2 + recall)
    ```

    This is the weighted harmonic mean of precision and recall.
    Its output range is `[0, 1]`. It operates at a word level
    and can be used for **QA systems**.

    If `y_true` and `y_pred` contains multiple fields
    The JSON object's fields are flattened and the score
    computed for each one independently.

    Args:
        average (str): Type of averaging to be performed across per-field results
            in the multi-field case.
            Acceptable values are `None`, `"micro"`, `"macro"` and
            `"weighted"`. Defaults to `None`.
            If `None`, no averaging is performed and `result()` will return
            the score for each class.
            If `"micro"`, compute metrics globally by counting the total
            true positives, false negatives and false positives.
            If `"macro"`, compute metrics for each label,
            and return their unweighted mean.
            This does not take label imbalance into account.
            If `"weighted"`, compute metrics for each label,
            and return their average weighted by support
            (the number of true instances for each label).
            This alters `"macro"` to account for label imbalance.
            It can result in an score that is not between precision and recall.
        beta (float): Determines the weight of given to recall
            in the harmonic mean between precision and recall (see pseudocode
            equation above). Defaults to `1`.
        name (str): (Optional) string name of the metric instance.
        in_mask (list): (Optional) list of keys to keep to compute the metric.
        out_mask (list): (Optional) list of keys to remove to compute the metric.
    """

    def __init__(
        self,
        average=None,
        beta=1.0,
        name="fbeta_score",
        in_mask=None,
        out_mask=None,
    ):
        super().__init__(
            name=name,
            in_mask=in_mask,
            out_mask=out_mask,
        )
        if average not in (None, "micro", "macro", "weighted"):
            raise ValueError(
                "Invalid `average` argument value. Expected one of: "
                "[None, 'micro', 'macro', 'weighted']. "
                f"Received: average={average}"
            )

        if not isinstance(beta, float):
            raise ValueError(
                "Invalid `beta` argument value. "
                "It should be a Python float. "
                f"Received: beta={beta} of type '{type(beta)}'"
            )
        self.state = self.add_variable(
            data_model=FBetaState,
            name=self.name + "_state",
        )
        self.average = average
        self.beta = beta
        self.axis = None
        if self.average != "micro":
            self.axis = 0

    async def update_state(self, y_true, y_pred):
        y_pred = tree.map_structure(lambda x: ops.convert_to_json_data_model(x), y_pred)
        y_true = tree.map_structure(lambda x: ops.convert_to_json_data_model(x), y_true)

        if self.in_mask:
            y_pred = tree.map_structure(lambda x: x.in_mask(mask=self.in_mask), y_pred)
            y_true = tree.map_structure(lambda x: x.in_mask(mask=self.in_mask), y_true)
        if self.out_mask:
            y_pred = tree.map_structure(lambda x: x.out_mask(mask=self.out_mask), y_pred)
            y_true = tree.map_structure(lambda x: x.out_mask(mask=self.out_mask), y_true)

        y_true = tree.flatten(tree.map_structure(lambda x: str(x), y_true.get_json()))
        y_pred = tree.flatten(tree.map_structure(lambda x: str(x), y_pred.get_json()))

        true_positives = []
        false_positives = []
        false_negatives = []
        intermediate_weights = []
        # For each field of y_true and y_pred
        for yt, yp in zip(y_true, y_pred):
            y_true_tokens = nlp_utils.normalize_and_tokenize(str(yt))
            y_pred_tokens = nlp_utils.normalize_and_tokenize(str(yp))
            common_tokens = set(y_true_tokens) & set(y_pred_tokens)
            true_positives.append(len(common_tokens))
            false_positives.append(len(y_pred_tokens) - len(common_tokens))
            false_negatives.append(len(y_true_tokens) - len(common_tokens))
            intermediate_weights.append(len(y_true_tokens))

        true_positives = np.convert_to_numpy(true_positives)
        false_positives = np.convert_to_numpy(false_positives)
        false_negatives = np.convert_to_numpy(false_negatives)
        intermediate_weights = np.convert_to_numpy(intermediate_weights)

        current_true_positives = self.state.get("true_positives")
        if current_true_positives:
            true_positives = np.add(current_true_positives, true_positives)

        current_false_positives = self.state.get("false_positives")
        if current_false_positives:
            false_positives = np.add(current_false_positives, false_positives)

        current_false_negatives = self.state.get("false_negatives")
        if current_false_negatives:
            false_negatives = np.add(current_false_negatives, false_negatives)

        current_intermediate_weights = self.state.get("intermediate_weights")
        if current_intermediate_weights:
            intermediate_weights = np.add(
                current_intermediate_weights, intermediate_weights
            )

        self.state.update(
            {
                "true_positives": true_positives.tolist(),
                "false_positives": false_positives.tolist(),
                "false_negatives": false_negatives.tolist(),
                "intermediate_weights": intermediate_weights.tolist(),
            }
        )

    def result(self):
        if (
            self.state.get("true_positives") is None
            and self.state.get("false_positives") is None
            and self.state.get("false_negatives") is None
        ):
            return 0.0
        precision = np.divide(
            self.state.get("true_positives"),
            np.add(
                self.state.get("true_positives"),
                self.state.get("false_positives"),
            )
            + backend.epsilon(),
        )
        recall = np.divide(
            self.state.get("true_positives"),
            np.add(
                self.state.get("true_positives"),
                self.state.get("false_negatives"),
            )
            + backend.epsilon(),
        )
        precision = np.convert_to_tensor(precision)
        recall = np.convert_to_tensor(recall)

        mul_value = precision * recall
        add_value = ((self.beta**2) * precision) + recall
        mean = np.divide(mul_value, add_value + backend.epsilon())
        f1_score = mean * (1 + (self.beta**2))
        if self.average == "weighted":
            intermediate_weights = self.state.get("intermediate_weights")
            weights = np.divide(
                intermediate_weights,
                np.sum(intermediate_weights) + backend.epsilon(),
            )
            f1_score = np.sum(f1_score * weights)

        elif self.average is not None:  # [micro, macro]
            f1_score = np.mean(f1_score, self.axis)

        try:
            return float(f1_score)
        except Exception:
            return list(f1_score)

    def get_config(self):
        """Return the serializable config of the metric.

        Returns:
            (dict): The config dict.
        """
        config = {
            "name": self.name,
            "beta": self.beta,
        }
        base_config = super().get_config()
        return {**base_config, **config}


@synalinks_export("synalinks.metrics.F1Score")
class F1Score(FBetaScore):
    """Computes F-1 Score.

    Formula:

    ```python
    f1_score = 2 * (precision * recall) / (precision + recall)
    ```

    This is the harmonic mean of precision and recall.
    Its output range is `[0, 1]`. It operates at a word level
    and can be used for **QA systems**.

    If `y_true` and `y_pred` contains multiple fields
    The JSON object's fields are flattened and the score
    computed for each one independently before being averaged.

    Args:
        average (str): Type of averaging to be performed across per-field results
            in the multi-field case.
            Acceptable values are `None`, `"micro"`, `"macro"` and
            `"weighted"`. Defaults to `None`.
            If `None`, no averaging is performed and `result()` will return
            the score for each class.
            If `"micro"`, compute metrics globally by counting the total
            true positives, false negatives and false positives.
            If `"macro"`, compute metrics for each label,
            and return their unweighted mean.
            This does not take label imbalance into account.
            If `"weighted"`, compute metrics for each label,
            and return their average weighted by support
            (the number of true instances for each label).
            This alters `"macro"` to account for label imbalance.
            It can result in an score that is not between precision and recall.
        name (str): (Optional) string name of the metric instance.
        in_mask (list): (Optional) list of keys to keep to compute the metric.
        out_mask (list): (Optional) list of keys to remove to compute the metric.
    """

    def __init__(
        self,
        average=None,
        name="f1_score",
        in_mask=None,
        out_mask=None,
    ):
        super().__init__(
            average=average,
            beta=1.0,
            name=name,
            in_mask=in_mask,
            out_mask=out_mask,
        )

    def get_config(self):
        """Return the serializable config of the metric.

        Returns:
            (dict): The config dict.
        """
        base_config = super().get_config()
        del base_config["beta"]
        return base_config


@synalinks_export("synalinks.metrics.BinaryFBetaScore")
class BinaryFBetaScore(FBetaScore):
    """Computes F-Beta score on binary structures.

    Formula:

    ```python
    b2 = beta ** 2
    f_beta_score = (1 + b2) * (precision * recall) / (precision * b2 + recall)
    ```

    This is the weighted harmonic mean of precision and recall.
    Its output range is `[0, 1]`. It operates at a field level
    and can be used for **multi-class and multi-label classification**.

    Each field of `y_true` and `y_pred` should be booleans or floats between [0, 1].
    If the fields are floats, it uses the threshold for deciding
    if the values are 0 or 1.

    Args:
        average (str): Type of averaging to be performed across per-class results
            in the multi-class case.
            Acceptable values are `None`, `"micro"`, `"macro"` and
            `"weighted"`. Defaults to `None`.
            If `None`, no averaging is performed and `result()` will return
            the score for each class.
            If `"micro"`, compute metrics globally by counting the total
            true positives, false negatives and false positives.
            If `"macro"`, compute metrics for each label,
            and return their unweighted mean.
            This does not take label imbalance into account.
            If `"weighted"`, compute metrics for each label,
            and return their average weighted by support
            (the number of true instances for each label).
            This alters `"macro"` to account for label imbalance.
            It can result in an score that is not between precision and recall.
        beta (float): Determines the weight of given to recall
            in the harmonic mean between precision and recall (see pseudocode
            equation above). Defaults to `1`.
        threshold (float): (Optional) Float representing the threshold for deciding
            whether prediction values are 1 or 0. Elements of `y_pred` and `y_true`
            greater than `threshold` are converted to be 1, and the rest 0.
        name (str): (Optional) string name of the metric instance.
        in_mask (list): (Optional) list of keys to keep to compute the metric.
        out_mask (list): (Optional) list of keys to remove to compute the metric.
    """

    def __init__(
        self,
        average=None,
        beta=1.0,
        threshold=0.5,
        name="binary_fbeta_score",
        in_mask=None,
        out_mask=None,
    ):
        super().__init__(
            average=average,
            beta=beta,
            name=name,
            in_mask=in_mask,
            out_mask=out_mask,
        )
        if not isinstance(threshold, float):
            raise ValueError(
                "Invalid `threshold` argument value. "
                "It should be a Python float. "
                f"Received: threshold={threshold} "
                f"of type '{type(threshold)}'"
            )
        if threshold > 1.0 or threshold <= 0.0:
            raise ValueError(
                "Invalid `threshold` argument value. "
                "It should verify 0 < threshold <= 1. "
                f"Received: threshold={threshold}"
            )
        self.threshold = threshold

    async def update_state(self, y_true, y_pred):
        y_pred = tree.map_structure(lambda x: ops.convert_to_json_data_model(x), y_pred)
        y_true = tree.map_structure(lambda x: ops.convert_to_json_data_model(x), y_true)

        if self.in_mask:
            y_pred = tree.map_structure(lambda x: x.in_mask(mask=self.in_mask), y_pred)
            y_true = tree.map_structure(lambda x: x.in_mask(mask=self.in_mask), y_true)
        if self.out_mask:
            y_pred = tree.map_structure(lambda x: x.out_mask(mask=self.out_mask), y_pred)
            y_true = tree.map_structure(lambda x: x.out_mask(mask=self.out_mask), y_true)

        def convert_to_binary(x):
            if isinstance(x, bool):
                return 1.0 if x is True else 0.0
            elif isinstance(x, float):
                return 1.0 if x > self.threshold else 0.0
            else:
                raise ValueError(
                    "All `y_true` and y_pred` fields should be booleans or floats. "
                    "Use `in_mask` or `out_mask` to remove the other fields."
                )

        y_true = tree.flatten(
            tree.map_structure(lambda x: convert_to_binary(x), y_true.get_json())
        )
        y_pred = tree.flatten(
            tree.map_structure(lambda x: convert_to_binary(x), y_pred.get_json())
        )
        y_true = np.convert_to_tensor(y_true)
        y_pred = np.convert_to_tensor(y_pred)

        true_positives = y_pred * y_true
        false_positives = y_pred * (1 - y_true)
        false_negatives = (1 - y_pred) * y_true
        intermediate_weights = y_true

        current_true_positives = self.state.get("true_positives")
        if current_true_positives:
            true_positives = np.add(current_true_positives, true_positives)

        current_false_positives = self.state.get("false_positives")
        if current_false_positives:
            false_positives = np.add(current_false_positives, false_positives)

        current_false_negatives = self.state.get("false_negatives")
        if current_false_negatives:
            false_negatives = np.add(current_false_negatives, false_negatives)

        current_intermediate_weights = self.state.get("intermediate_weights")
        if current_intermediate_weights:
            intermediate_weights = np.add(
                current_intermediate_weights, intermediate_weights
            )

        self.state.update(
            {
                "true_positives": true_positives.tolist(),
                "false_positives": false_positives.tolist(),
                "false_negatives": false_negatives.tolist(),
                "intermediate_weights": intermediate_weights.tolist(),
            }
        )

    def get_config(self):
        """Return the serializable config of the metric.

        Returns:
            (dict): The config dict.
        """
        config = {
            "beta": self.beta,
            "threshold": self.threshold,
            "name": self.name,
        }
        base_config = super().get_config()
        return {**base_config, **config}


@synalinks_export("synalinks.metrics.BinaryF1Score")
class BinaryF1Score(BinaryFBetaScore):
    """Computes F-1 Score on binary structures.

    Formula:

    ```python
    f1_score = 2 * (precision * recall) / (precision + recall)
    ```

    This is the harmonic mean of precision and recall.
    Its output range is `[0, 1]`. It operates at a field level
    and can be used for **multi-class and multi-label classification**.

    Each field of `y_true` and `y_pred` should booleans or floats between [0, 1].
    If the fields are floats, it uses the threshold for deciding
    if the values are 0 or 1.

    Args:
        average (str): Type of averaging to be performed across per-class results
            in the multi-class case.
            Acceptable values are `None`, `"micro"`, `"macro"` and
            `"weighted"`. Defaults to `None`.
            If `None`, no averaging is performed and `result()` will return
            the score for each class.
            If `"micro"`, compute metrics globally by counting the total
            true positives, false negatives and false positives.
            If `"macro"`, compute metrics for each label,
            and return their unweighted mean.
            This does not take label imbalance into account.
            If `"weighted"`, compute metrics for each label,
            and return their average weighted by support
            (the number of true instances for each label).
            This alters `"macro"` to account for label imbalance.
            It can result in an score that is not between precision and recall.
        threshold (float): (Optional) Float representing the threshold for deciding
            whether prediction values are 1 or 0. Elements of `y_pred` and `y_true`
            greater than `threshold` are converted to be 1, and the rest 0.
        name (str): (Optional) string name of the metric instance.
        in_mask (list): (Optional) list of keys to keep to compute the metric.
        out_mask (list): (Optional) list of keys to remove to compute the metric.
    """

    def __init__(
        self,
        average=None,
        threshold=0.5,
        name="binary_f1_score",
        in_mask=None,
        out_mask=None,
    ):
        super().__init__(
            average=average,
            beta=1.0,
            threshold=threshold,
            name=name,
            in_mask=in_mask,
            out_mask=out_mask,
        )

    def get_config(self):
        """Return the serializable config of the metric.

        Returns:
            (dict): The config dict.
        """
        base_config = super().get_config()
        del base_config["beta"]
        return base_config



# # License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

# TODO check why the CI/CD fail, maybe network issue with MemGraph ?!

# from typing import Literal
# from typing import Union
# from unittest.mock import patch

# import numpy as np

# from synalinks.src import testing
# from synalinks.src.backend import Entity
# from synalinks.src.backend import Relation
# from synalinks.src.backend import SimilaritySearch
# from synalinks.src.embedding_models import EmbeddingModel
# from synalinks.src.knowledge_bases.database_adapters.memgraph_adapter import (
#     MemGraphAdapter,
# )
# from synalinks.src.modules import Embedding
# from synalinks.src.modules import Input
# from synalinks.src.programs import Program


# class Document(Entity):
#     label: Literal["Document"]
#     text: str


# class Chunk(Document):
#     label: Literal["Chunk"]
#     text: str


# class IsPartOf(Relation):
#     obj: Union[Chunk, Document]
#     label: Literal["IsPartOf"]
#     subj: Document


# class MemGraphAdapterTest(testing.TestCase):
#     @patch("litellm.aembedding")
#     async def test_adapter(self, mock_embedding):
#         expected_value = np.random.rand(1024)
#         mock_embedding.return_value = {"data": [{"embedding": expected_value}]}

#         embedding_model = EmbeddingModel(
#             model="ollama/mxbai-embed-large",
#         )

#         adapter = MemGraphAdapter(
#             uri="memgraph://localhost:7688",
#             embedding_model=embedding_model,
#             entity_models=[Document, Chunk],
#             relation_models=[IsPartOf],
#             wipe_on_start=True,
#         )

#         _ = await adapter.query("RETURN 1")

#     @patch("litellm.aembedding")
#     async def test_adapter_update_entity(self, mock_embedding):
#         expected_value = np.random.rand(1024)
#         mock_embedding.return_value = {"data": [{"embedding": expected_value}]}

#         embedding_model = EmbeddingModel(model="ollama/mxbai-embed-large")

#         adapter = MemGraphAdapter(
#             uri="memgraph://localhost:7688",
#             embedding_model=embedding_model,
#             entity_models=[Document, Chunk],
#             relation_models=[IsPartOf],
#             wipe_on_start=True,
#         )

#         doc1 = Document(
#             label="Document",
#             text="test document 1",
#         )

#         inputs = Input(data_model=Document)
#         outputs = await Embedding(
#             embedding_model=embedding_model,
#             in_mask=["text"],
#         )(inputs)

#         program = Program(
#             inputs=inputs,
#             outputs=outputs,
#         )

#         embedded_doc = await program(doc1)

#         await adapter.update(embedded_doc)

#     @patch("litellm.aembedding")
#     async def test_adapter_update_relation(self, mock_embedding):
#         expected_value = np.random.rand(1024)
#         mock_embedding.return_value = {"data": [{"embedding": expected_value}]}

#         embedding_model = EmbeddingModel(
#             model="ollama/mxbai-embed-large",
#         )

#         adapter = MemGraphAdapter(
#             uri="memgraph://localhost:7688",
#             embedding_model=embedding_model,
#             entity_models=[Document, Chunk],
#             relation_models=[IsPartOf],
#             wipe_on_start=True,
#         )

#         doc1 = Document(
#             label="Document",
#             text="test document 1",
#         )
#         doc2 = Document(
#             label="Document",
#             text="test document 2",
#         )

#         rel1 = IsPartOf(
#             subj=doc1,
#             label="IsPartOf",
#             obj=doc2,
#         )

#         inputs = Input(data_model=Document)
#         outputs = await Embedding(
#             embedding_model=embedding_model,
#             in_mask=["text"],
#         )(inputs)

#         program = Program(
#             inputs=inputs,
#             outputs=outputs,
#         )

#         embedded_rel = await program(rel1)

#         await adapter.update(embedded_rel)

#     @patch("litellm.aembedding")
#     async def test_adapter_similarity_search(self, mock_embedding):
#         expected_value = np.random.rand(1024)
#         mock_embedding.return_value = {"data": [{"embedding": expected_value}]}

#         embedding_model = EmbeddingModel(model="ollama/mxbai-embed-large")

#         adapter = MemGraphAdapter(
#             uri="memgraph://localhost:7688",
#             embedding_model=embedding_model,
#             entity_models=[Document, Chunk],
#             relation_models=[IsPartOf],
#             wipe_on_start=True,
#         )

#         inputs = Input(data_model=Document)

#         outputs = await Embedding(
#             embedding_model=embedding_model,
#             in_mask=["text"],
#         )(inputs)

#         program = Program(
#             inputs=inputs,
#             outputs=outputs,
#         )

#         doc1 = Document(
#             label="Document",
#             text="test document 1",
#         )
#         doc2 = Document(
#             label="Document",
#             text="test document 2",
#         )
#         doc3 = Document(
#             label="Document",
#             text="test document 3",
#         )

#         batch = np.array([doc1, doc2, doc3], dtype=object)

#         embedded_docs = await program.predict(batch)

#         await adapter.update(embedded_docs[0])
#         await adapter.update(embedded_docs[1])
#         await adapter.update(embedded_docs[2])

#         search = SimilaritySearch(
#             entity_label="Document",
#             similarity_search="test document",
#         ).to_json_data_model()
#         result = await adapter.similarity_search(search, threshold=0.0)
#         self.assertTrue(len(result) > 0)

#     @patch("litellm.aembedding")
#     async def test_adapter_triplet_search_basic(self, mock_embedding):
#         """Test basic triplet search functionality"""
#         expected_value = np.random.rand(1024)
#         mock_embedding.return_value = {"data": [{"embedding": expected_value}]}

#         embedding_model = EmbeddingModel(model="ollama/mxbai-embed-large")

#         adapter = MemGraphAdapter(
#             uri="memgraph://localhost:7688",
#             embedding_model=embedding_model,
#             entity_models=[Document, Chunk],
#             relation_models=[IsPartOf],
#             wipe_on_start=True,
#         )

#         # Create test documents
#         doc1 = Document(label="Document", text="test document 1")
#         doc2 = Document(label="Document", text="test document 2")
#         chunk1 = Chunk(label="Chunk", text="test chunk 1")

#         # Create embeddings
#         inputs = Input(data_model=Document)
#         outputs = await Embedding(
#             embedding_model=embedding_model,
#             in_mask=["text"],
#         )(inputs)
#         program = Program(inputs=inputs, outputs=outputs)

#         # Embed and store entities
#         embedded_doc1 = await program(doc1)
#         embedded_doc2 = await program(doc2)
#         embedded_chunk1 = await program(chunk1)

#         await adapter.update(embedded_doc1)
#         await adapter.update(embedded_doc2)
#         await adapter.update(embedded_chunk1)

#         # Create and store relation
#         rel1 = IsPartOf(subj=chunk1, label="IsPartOf", obj=doc1)
#         embedded_rel = await program(rel1)
#         await adapter.update(embedded_rel)

#         # Test triplet search
#         from synalinks.src.backend import TripletSearch

#         triplet_search = TripletSearch(
#             subject_label="Chunk",
#             subject_similarity_search="",
#             relation_label="IsPartOf",
#             object_label="Document",
#             object_similarity_search="",
#         ).to_json_data_model()

#         result = await adapter.triplet_search(triplet_search, threshold=0.0)
#         self.assertTrue(len(result) > 0)
#         self.assertIn("subj", result[0])
#         self.assertIn("relation", result[0])
#         self.assertIn("obj", result[0])

#     @patch("litellm.aembedding")
#     async def test_adapter_triplet_search_with_similarity(self, mock_embedding):
#         """Test triplet search with similarity queries"""
#         expected_value = np.random.rand(1024)
#         mock_embedding.return_value = {"data": [{"embedding": expected_value}]}

#         embedding_model = EmbeddingModel(model="ollama/mxbai-embed-large")

#         adapter = MemGraphAdapter(
#             uri="memgraph://localhost:7688",
#             embedding_model=embedding_model,
#             entity_models=[Document, Chunk],
#             relation_models=[IsPartOf],
#             wipe_on_start=True,
#         )

#         # Create and embed test data
#         doc1 = Document(label="Document", text="artificial intelligence document")
#         chunk1 = Chunk(label="Chunk", text="machine learning chapter")

#         inputs = Input(data_model=Document)
#         outputs = await Embedding(
#             embedding_model=embedding_model,
#             in_mask=["text"],
#         )(inputs)
#         program = Program(inputs=inputs, outputs=outputs)

#         embedded_doc1 = await program(doc1)
#         embedded_chunk1 = await program(chunk1)

#         await adapter.update(embedded_doc1)
#         await adapter.update(embedded_chunk1)

#         # Create relation
#         rel1 = IsPartOf(subj=chunk1, label="IsPartOf", obj=doc1)
#         embedded_rel = await program(rel1)
#         await adapter.update(embedded_rel)

#         # Test triplet search with subject similarity
#         from synalinks.src.backend import TripletSearch

#         triplet_search = TripletSearch(
#             subject_label="Chunk",
#             subject_similarity_search="machine learning",
#             relation_label="IsPartOf",
#             object_label="Document",
#             object_similarity_search="?",
#         ).to_json_data_model()

#         result = await adapter.triplet_search(triplet_search, threshold=0.0)
#         self.assertTrue(len(result) > 0)

#     @patch("litellm.aembedding")
#     async def test_adapter_triplet_search_with_object_similarity(self, mock_embedding):
#         """Test triplet search with object similarity query"""
#         expected_value = np.random.rand(1024)
#         mock_embedding.return_value = {"data": [{"embedding": expected_value}]}

#         embedding_model = EmbeddingModel(model="ollama/mxbai-embed-large")

#         adapter = MemGraphAdapter(
#             uri="memgraph://localhost:7688",
#             embedding_model=embedding_model,
#             entity_models=[Document, Chunk],
#             relation_models=[IsPartOf],
#             wipe_on_start=True,
#         )

#         # Create and embed test data
#         doc1 = Document(label="Document", text="programming tutorial")
#         chunk1 = Chunk(label="Chunk", text="python basics")

#         inputs = Input(data_model=Document)
#         outputs = await Embedding(
#             embedding_model=embedding_model,
#             in_mask=["text"],
#         )(inputs)
#         program = Program(inputs=inputs, outputs=outputs)

#         embedded_doc1 = await program(doc1)
#         embedded_chunk1 = await program(chunk1)

#         await adapter.update(embedded_doc1)
#         await adapter.update(embedded_chunk1)

#         # Create relation
#         rel1 = IsPartOf(subj=chunk1, label="IsPartOf", obj=doc1)
#         embedded_rel = await program(rel1)
#         await adapter.update(embedded_rel)

#         # Test triplet search with object similarity
#         from synalinks.src.backend import TripletSearch

#         triplet_search = TripletSearch(
#             subject_label="Chunk",
#             subject_similarity_search="?",
#             relation_label="IsPartOf",
#             object_label="Document",
#             object_similarity_search="programming tutorial",
#         ).to_json_data_model()

#         result = await adapter.triplet_search(triplet_search, threshold=0.0)
#         self.assertTrue(len(result) > 0)

#     @patch("litellm.aembedding")
#     async def test_adapter_triplet_search_with_both_similarities(self, mock_embedding):
#         """Test triplet search with both subject and object similarity queries"""
#         expected_value = np.random.rand(1024)
#         mock_embedding.return_value = {"data": [{"embedding": expected_value}]}

#         embedding_model = EmbeddingModel(model="ollama/mxbai-embed-large")

#         adapter = MemGraphAdapter(
#             uri="memgraph://localhost:7688",
#             embedding_model=embedding_model,
#             entity_models=[Document, Chunk],
#             relation_models=[IsPartOf],
#             wipe_on_start=True,
#         )

#         # Create test data
#         doc1 = Document(label="Document", text="data science handbook")
#         doc2 = Document(label="Document", text="machine learning guide")
#         chunk1 = Chunk(label="Chunk", text="statistics chapter")
#         chunk2 = Chunk(label="Chunk", text="neural networks section")

#         inputs = Input(data_model=Document)
#         outputs = await Embedding(
#             embedding_model=embedding_model,
#             in_mask=["text"],
#         )(inputs)
#         program = Program(inputs=inputs, outputs=outputs)

#         # Embed and store all entities
#         entities = [doc1, doc2, chunk1, chunk2]
#         for entity in entities:
#             embedded_entity = await program(entity)
#             await adapter.update(embedded_entity)

#         # Create relations
#         rel1 = IsPartOf(subj=chunk1, label="IsPartOf", obj=doc1)
#         rel2 = IsPartOf(subj=chunk2, label="IsPartOf", obj=doc2)

#         for rel in [rel1, rel2]:
#             embedded_rel = await program(rel)
#             await adapter.update(embedded_rel)

#         # Test triplet search with both similarities
#         from synalinks.src.backend import TripletSearch

#         triplet_search = TripletSearch(
#             subject_label="Chunk",
#             subject_similarity_search="statistics",
#             relation_label="IsPartOf",
#             object_label="Document",
#             object_similarity_search="data science",
#         ).to_json_data_model()

#         result = await adapter.triplet_search(triplet_search, threshold=0.0)
#         self.assertTrue(len(result) > 0)



# Modified from: keras/src/backend/numpy/numpy.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

# Because we only use tensors for metrics and rewards, we don't need much
# and because we don't perform any gradient descent
# nor store weights that can benefit from specific dtype
# we don't need dtype inference as we can use floatx for everything

import numpy as np

from synalinks.src.backend import floatx


def standardize_axis_for_numpy(axis):
    """Standardize an axis to a tuple if it is a list."""
    return tuple(axis) if isinstance(axis, list) else axis


def convert_to_numpy(x):
    return np.array(x)


def shape(x):
    return x.shape


def convert_to_tensor(x):
    return np.array(x, dtype=floatx())


def zeros(shape, dtype=None):
    return np.zeros(shape, dtype=floatx())


def add(x1, x2):
    x1 = convert_to_tensor(x1)
    x2 = convert_to_tensor(x2)
    return np.add(x1, x2)


def subtract(x1, x2):
    x1 = convert_to_tensor(x1)
    x2 = convert_to_tensor(x2)
    return np.subtract(x1, x2)


def multiply(x1, x2):
    x1 = convert_to_tensor(x1)
    x2 = convert_to_tensor(x2)
    return np.multiply(x1, x2)


def mean(x, axis=None, keepdims=False):
    axis = standardize_axis_for_numpy(axis)
    x = convert_to_tensor(x)
    return np.mean(x, axis=axis, keepdims=keepdims).astype(floatx())


def median(x, axis=None, keepdims=False):
    axis = standardize_axis_for_numpy(axis)
    x = convert_to_tensor(x)
    return np.median(x, axis=axis, keepdims=keepdims).astype(floatx())


def sum(x, axis=None, keepdims=False):
    axis = standardize_axis_for_numpy(axis)
    return np.sum(x, axis=axis, keepdims=keepdims).astype(floatx())


def divide(x1, x2):
    x1 = convert_to_tensor(x1)
    x2 = convert_to_tensor(x2)
    return np.divide(x1, x2)


def prod(x, axis=None, keepdims=False):
    axis = standardize_axis_for_numpy(axis)
    x = convert_to_tensor(x)
    return np.prod(x, axis=axis, keepdims=keepdims, dtype=floatx())


def squeeze(x, axis=None):
    axis = standardize_axis_for_numpy(axis)
    return np.squeeze(x, axis=axis)


def expand_dims(x, axis):
    axis = standardize_axis_for_numpy(axis)
    return np.expand_dims(x, axis)


def divide_no_nan(x1, x2):
    x1 = convert_to_tensor(x1)
    x2 = convert_to_tensor(x2)
    # No need for the double-where trick since we don't calculate gradients in
    # numpy backend.
    return np.where(x2 == 0, np.array(0, dtype=floatx()), np.divide(x1, x2))


def broadcast_to(x, shape):
    return np.broadcast_to(x, shape)


def normalize(x, axis=-1, order=2):
    norm = np.atleast_1d(np.linalg.norm(x, order, axis))
    norm[norm == 0] = 1

    # axis cannot be `None`
    axis = axis or -1
    return x / np.expand_dims(norm, axis)



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.datasets.hotpotqa import get_input_data_model as get_input_data_model
from synalinks.src.datasets.hotpotqa import (
    get_knowledge_data_model as get_knowledge_data_model,
)
from synalinks.src.datasets.hotpotqa import get_output_data_model as get_output_data_model
from synalinks.src.datasets.hotpotqa import load_data as load_data
from synalinks.src.datasets.hotpotqa import load_knowledge as load_knowledge



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import utils
from synalinks.src.api_export import synalinks_export


@synalinks_export("synalinks.hooks.Hook")
class Hook:
    """Base hook class used to build new hooks.

    Hooks are callback-like objects that can be passed to the module's configuration, they
    intercept the `__call__()` method of the modules.

    They can be used to log the inputs/outputs of the modules as well as for
    streaming data or for observability.

    Attributes:
        params (dict): Hook parameters
            (eg. endpoint, log-level ...).
        module (Module): Instance of `Module`.
            Reference of the module being monitored.
    """

    def __init__(self):
        self.params = None
        self._module = None

    def set_params(self, params):
        self.params = params

    def set_module(self, module):
        self._module = module

    @property
    def module(self):
        return self._module

    @utils.default
    def on_call_begin(
        self,
        call_id,
        inputs=None,
    ):
        """Called at the beginning of the module execution.

        Args:
            call_id (str): The id of the module's call
            inputs (SymbolicDataModel | JsonDataModel | DataModel | list | dict | tuple):
                The module's inputs. The outputs can be data models or lists,
                dicts or tuples of data models.
        """
        pass

    @utils.default
    def on_call_end(
        self,
        call_id,
        outputs=None,
        exception=None,
    ):
        """Called at the end of the module execution.

        Args:
            call_id (str): The id of the module's call
            outputs (SymbolicDataModel | JsonDataModel | DataModel | list | dict | tuple):
                The module's outputs. The outputs can be data models or lists,
                dicts or tuples of data models.
            exception (str): Exception message if any.
        """
        pass



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from unittest.mock import patch

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.embedding_models import EmbeddingModel
from synalinks.src.rewards.cosine_similarity import CosineSimilarity


class CosineSimilarityTest(testing.TestCase):
    @patch("litellm.aembedding")
    async def test_base_function(self, mock_embedding):
        embedding_model = EmbeddingModel(model="ollama/all-minilm")
        expected_value = [0.0, 0.1, 0.2, 0.3, 0.4]
        mock_embedding.return_value = {"data": [{"embedding": expected_value}]}

        class Answer(DataModel):
            answer: str

        y_true = Answer(answer="Paris")
        y_pred = Answer(answer="Paris")

        cosine_similarity = CosineSimilarity(embedding_model=embedding_model)
        reward = await cosine_similarity(y_true, y_pred)
        self.assertEqual(reward, 1.0)

    @patch("litellm.aembedding")
    async def test_multiple_fields(self, mock_embedding):
        embedding_model = EmbeddingModel(model="ollama/all-minilm")
        expected_value = [0.0, 0.1, 0.2, 0.3, 0.4]
        mock_embedding.return_value = {
            "data": [{"embedding": expected_value}, {"embedding": expected_value}]
        }

        class Answer(DataModel):
            thinking: str
            answer: str

        y_true = Answer(thinking="The capital of France is paris", answer="Paris")
        y_pred = Answer(thinking="The capital of France is paris", answer="Paris")

        cosine_similarity = CosineSimilarity(embedding_model=embedding_model)
        reward = await cosine_similarity(y_true, y_pred)
        self.assertEqual(reward, 1.0)



# Modified from: keras/src/trainers/data_adapters/data_adapter_utils.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src.api_export import synalinks_export

NUM_BATCHES_FOR_SPEC = 2


@synalinks_export("synalinks.utils.unpack_x_y")
def unpack_x_y(data):
    """Unpacks user-provided data tuple.

    This is a convenience utility to be used when overriding
    `Program.train_step`, `Program.test_step`, or `Program.predict_step`.
    This utility makes it easy to support data of the form `(x,)`, or
    `(x, y)`.

    Args:
        data: A tuple of the form `(x,)`, or `(x, y)`.

    Returns:
        The unpacked tuple, with `None`s for `y` if they are
        not provided.
    """
    if isinstance(data, list):
        data = tuple(data)
    if not isinstance(data, tuple):
        return (data, None)
    elif len(data) == 1:
        return (data[0], None)
    elif len(data) == 2:
        return (data[0], data[1])
    error_msg = (
        f"Data is expected to be in format `x`, `(x,)`, or `(x, y)`, found: {data}"
    )
    raise ValueError(error_msg)


@synalinks_export("synalinks.utils.pack_x_y")
def pack_x_y(x, y=None):
    """Packs user-provided data into a tuple.

    This is a convenience utility for packing data into the tuple formats
    that `Program.fit()` uses.

    Args:
        x: `DataModel`s to pass to `Program`.
        y: Ground-truth targets to pass to `Program`.

    Returns:
        Tuple in the format used in `Program.fit()`.
    """
    if y is None:
        return (x,)
    return (x, y)



import inspect

from synalinks.src import backend
from synalinks.src.api_export import synalinks_export
from synalinks.src.initializers.empty_initializer import Empty
from synalinks.src.initializers.initializer import Initializer
from synalinks.src.saving import serialization_lib
from synalinks.src.utils.naming import to_snake_case

ALL_OBJECTS = {
    Initializer,
    Empty,
}

ALL_OBJECTS_DICT = {cls.__name__: cls for cls in ALL_OBJECTS}
ALL_OBJECTS_DICT.update({to_snake_case(cls.__name__): cls for cls in ALL_OBJECTS})


@synalinks_export("synalinks.initializers.serialize")
def serialize(initializer):
    """Returns the initializer configuration as a Python dict."""
    return serialization_lib.serialize_synalinks_object(initializer)


@synalinks_export("synalinks.initializers.deserialize")
def deserialize(config, custom_objects=None):
    """Returns a Synalinks initializer object via its configuration."""
    return serialization_lib.deserialize_synalinks_object(
        config,
        module_objects=ALL_OBJECTS_DICT,
        custom_objects=custom_objects,
    )


@synalinks_export("synalinks.initializers.get")
def get(identifier):
    """Retrieves a Syanlinks initializer object via an identifier.

    The `identifier` may be the string name of a initializers function or class
    (case-sensitively).

    >>> identifier = 'Empty'
    >>> synalinks.initializers.get(identifier)
    <...synalinks.initializers.Empty...>

    You can also specify `config` of the initializer to this function by passing
    dict containing `class_name` and `config` as an identifier. Also note that
    the `class_name` must map to a `Initializer` class.

    >>> cfg = {'class_name': 'Empty', 'config': {...}}
    >>> synalinks.initializers.get(cfg)
    <...synalinks.initializers.Empty...>

    In the case that the `identifier` is a class, this method will return a new
    instance of the class by its constructor.

    Args:
        identifier (str | dict): A string, or dict specifying
            the initializer. If a string, it should be the name of an
            initializer. If a dict, it should contain the configuration of an
            initializer.

    Returns:
        Initializer instance base on the input identifier.
    """
    if identifier is None:
        return None
    if isinstance(identifier, dict):
        obj = deserialize(identifier)
    elif isinstance(identifier, str):
        config = {"class_name": str(identifier), "config": {}}
        obj = deserialize(config)
    else:
        obj = identifier
    if callable(obj):
        if inspect.isclass(obj):
            obj = obj()
        return obj
    else:
        raise ValueError(f"Could not interpret initializer identifier: {identifier}")



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import is_meta_class
from synalinks.src.initializers.initializer import Initializer


@synalinks_export(
    [
        "synalinks.initializers.Empty",
        "synalinks.initializers.empty",
    ]
)
class Empty(Initializer):
    """
    Initialize a variable with the data_model default value.

    Args:
        schema (dict): The JSON object schema. If not provided,
            uses the `data_model` to infer it.
        value (dict): The initial JSON object. If the value is not provided the
            `data_model` should be provided. When a value is provided, a schema
            should be provided.
        data_model (DataModel): The backend data_model to use.

    """

    def get_config(self):
        return {
            "schema": self._schema,
            "json": self._json,
        }

    def __call__(self, data_model=None):
        """Returns a JSON object initialized as specified by the initializer."""
        if data_model:
            if not is_meta_class(data_model):
                self._json = data_model.get_json()
            else:
                self._json = data_model().get_json()
            self._schema = data_model.get_schema()
        return self._json



# Modified from: keras/src/metrics/reduction_metrics.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import ops
from synalinks.src import rewards
from synalinks.src import tree
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import DataModel
from synalinks.src.backend.common import numpy
from synalinks.src.metrics.metric import Metric
from synalinks.src.saving import serialization_lib


def reduce_to_samplewise_values(values, reduce_fn):
    values = numpy.convert_to_tensor(values)
    values_ndim = len(values.shape)
    if values_ndim > 1:
        values = reduce_fn(values, axis=list(range(1, values_ndim)))
    return values


class Total(DataModel):
    total: float = 0.0


@synalinks_export("synalinks.metrics.Sum")
class Sum(Metric):
    """Compute the (weighted) sum of the given values.

    For example, if `values` is `[1, 3, 5, 7]` then their sum is 16.

    This metric creates one variable, `total`.
    This is ultimately returned as the sum value.

    Args:
        name (str): (Optional) string name of the metric instance.
        in_mask (list): (Optional) list of keys to keep to compute the metric.
        out_mask (list): (Optional) list of keys to remove to compute the metric.

    Example:

    ```python
    >>> m = metrics.Sum()
    >>> m.update_state([1, 3, 5, 7])
    >>> m.result()
    16.0
    ```
    """

    def __init__(self, name="sum", in_mask=None, out_mask=None):
        super().__init__(name=name, in_mask=in_mask, out_mask=out_mask)
        self.total = self.add_variable(
            data_model=Total,
            name="total",
        )

    async def update_state(self, values):
        values = reduce_to_samplewise_values(values, reduce_fn=numpy.sum)
        total = self.total.get("total")
        self.total.update({"total": float(numpy.sum(total, values))})

    def reset_state(self):
        self.total.assign(Total())

    def result(self):
        return self.total.get("total")


class TotalWithCount(DataModel):
    total: float = 0.0
    count: int = 0


@synalinks_export("synalinks.metrics.Mean")
class Mean(Metric):
    """Compute the mean of the given values.

    For example, if values is `[1, 3, 5, 7]` then the mean is 4.

    This metric creates two variables, `total` and `count`.
    The mean value returned is simply `total` divided by `count`.

    Args:
        name (str): (Optional) string name of the metric instance.
        in_mask (list): (Optional) list of keys to keep to compute the metric.
        out_mask (list): (Optional) list of keys to remove to compute the metric.

    Example:

    ```python
    >>> m = Mean()
    >>> m.update_state([1, 3, 5, 7])
    >>> m.result()
    4.0
    ```
    """

    def __init__(self, name="mean", in_mask=None, out_mask=None):
        super().__init__(name=name, in_mask=in_mask, out_mask=out_mask)
        self.total_with_count = self.add_variable(
            data_model=TotalWithCount, name="total_with_count"
        )

    async def update_state(self, values):
        values = reduce_to_samplewise_values(values, reduce_fn=numpy.mean)
        total = self.total_with_count.get("total")
        self.total_with_count.update({"total": float(total + numpy.sum(values))})
        if len(values.shape) >= 1:
            num_samples = numpy.shape(values)[0]
        else:
            num_samples = 1
        count = self.total_with_count.get("count")
        self.total_with_count.update({"count": int(count + num_samples)})

    def reset_state(self):
        self.total_with_count.assign(TotalWithCount())

    def result(self):
        return float(
            numpy.divide_no_nan(
                self.total_with_count.get("total"),
                self.total_with_count.get("count"),
            )
        )


@synalinks_export("synalinks.metrics.MeanMetricWrapper")
class MeanMetricWrapper(Mean):
    """Wrap a stateless metric function with the `Mean` metric.

    You could use this class to quickly build a mean metric from a function. The
    function needs to have the signature `fn(y_true, y_pred)` and return a
    per-sample reward array. `MeanMetricWrapper.result()` will return
    the average metric value across all samples seen so far.

    Args:
        fn (callable): The metric function to wrap, with signature
            `fn(y_true, y_pred, **kwargs)`.
        name (str): (Optional) string name of the metric instance.
        in_mask (list): (Optional) list of keys to keep to compute the metric.
        out_mask (list): (Optional) list of keys to remove to compute the metric.
        **kwargs (keyword arguments): Keyword arguments to pass on to `fn`.
    """

    def __init__(self, fn, name=None, in_mask=None, out_mask=None, **kwargs):
        super().__init__(
            name=name,
            in_mask=in_mask,
            out_mask=out_mask,
        )
        self._fn = fn
        self._fn_kwargs = kwargs

        # If we are wrapping a Synalinks reward, register the metric's
        # direction as "up" (needs to be maximized during training).
        if (
            self._fn in rewards.ALL_OBJECTS
            or hasattr(self._fn, "__class__")
            and self._fn.__class__ in rewards.ALL_OBJECTS
        ):
            self._direction = "up"

    async def update_state(self, y_true, y_pred):
        y_pred = tree.map_structure(lambda x: ops.convert_to_json_data_model(x), y_pred)
        y_true = tree.map_structure(lambda x: ops.convert_to_json_data_model(x), y_true)
        if self.in_mask:
            y_pred = tree.map_structure(lambda x: x.in_mask(mask=self.in_mask), y_pred)
            y_true = tree.map_structure(lambda x: x.in_mask(mask=self.in_mask), y_true)
        if self.out_mask:
            y_pred = tree.map_structure(lambda x: x.out_mask(mask=self.out_mask), y_pred)
            y_true = tree.map_structure(lambda x: x.out_mask(mask=self.out_mask), y_true)
        values = await self._fn(y_true, y_pred, **self._fn_kwargs)
        return await super().update_state(values)

    def get_config(self):
        """Returns the serializable config of the metric."""
        base_config = super().get_config()
        config = {
            "fn": serialization_lib.serialize_synalinks_object(self._fn),
        }
        config.update(self._fn_kwargs)
        return {**base_config, **config}

    @classmethod
    def from_config(cls, config):
        fn = serialization_lib.deserialize_synalinks_object(config.pop("fn"))
        return cls(fn=fn, **config)



import json
import os

from synalinks.cli.constants import PROJECT_CONFIG_FILENAME


def is_inside_synalinks_project():
    """
    Check if the current directory contains a synalinks project config.

    Returns:
        (bool): True if a synalinks project is detected, False otherwise.
    """
    if os.path.exists(os.path.join(os.getcwd(), PROJECT_CONFIG_FILENAME)):
        return True
    return False


def get_synalinks_project_config():
    """
    Retrieve the configuration of the current synalinks project.

    Returns:
        (dict): The project configuration if inside a synalinks project,
            otherwise an empty dictionary.
    """
    if is_inside_synalinks_project():
        with open(
            os.path.join(os.getcwd(), PROJECT_CONFIG_FILENAME, "r", encoding="utf-8")
        ) as f:
            return json.loads(f.read())
    return {}



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.backend.common.global_state import clear_session as clear_session
from synalinks.src.backend.common.json_data_model import (
    is_json_data_model as is_json_data_model,
)
from synalinks.src.backend.common.symbolic_data_model import (
    is_symbolic_data_model as is_symbolic_data_model,
)
from synalinks.src.backend.pydantic.core import is_meta_class as is_meta_class
from synalinks.src.saving.object_registration import (
    CustomObjectScope as CustomObjectScope,
)
from synalinks.src.saving.object_registration import (
    CustomObjectScope as custom_object_scope,
)
from synalinks.src.saving.object_registration import (
    get_custom_objects as get_custom_objects,
)
from synalinks.src.saving.object_registration import (
    get_registered_name as get_registered_name,
)
from synalinks.src.saving.object_registration import (
    get_registered_object as get_registered_object,
)
from synalinks.src.saving.object_registration import (
    register_synalinks_serializable as register_synalinks_serializable,
)
from synalinks.src.saving.serialization_lib import (
    deserialize_synalinks_object as deserialize_synalinks_object,
)
from synalinks.src.saving.serialization_lib import (
    serialize_synalinks_object as serialize_synalinks_object,
)
from synalinks.src.trainers.data_adapters.data_adapter_utils import pack_x_y as pack_x_y
from synalinks.src.trainers.data_adapters.data_adapter_utils import (
    unpack_x_y as unpack_x_y,
)
from synalinks.src.utils.file_utils import get_file as get_file
from synalinks.src.utils.io_utils import (
    disable_interactive_logging as disable_interactive_logging,
)
from synalinks.src.utils.io_utils import (
    enable_interactive_logging as enable_interactive_logging,
)
from synalinks.src.utils.io_utils import (
    is_interactive_logging_enabled as is_interactive_logging_enabled,
)
from synalinks.src.utils.plot_history import plot_history as plot_history
from synalinks.src.utils.plot_history import (
    plot_history_comparison as plot_history_comparison,
)
from synalinks.src.utils.plot_history import (
    plot_history_comparison_with_mean_and_std as plot_history_comparison_with_mean_and_std,
)
from synalinks.src.utils.plot_history import (
    plot_history_with_mean_and_std as plot_history_with_mean_and_std,
)
from synalinks.src.utils.plot_metrics import plot_metrics as plot_metrics
from synalinks.src.utils.plot_metrics import (
    plot_metrics_comparison as plot_metrics_comparison,
)
from synalinks.src.utils.plot_metrics import (
    plot_metrics_comparison_with_mean_and_std as plot_metrics_comparison_with_mean_and_std,
)
from synalinks.src.utils.plot_metrics import (
    plot_metrics_with_mean_and_std as plot_metrics_with_mean_and_std,
)
from synalinks.src.utils.progbar import Progbar as Progbar
from synalinks.src.utils.program_visualization import plot_program as plot_program
from synalinks.src.utils.program_visualization import program_to_dot as program_to_dot
from synalinks.src.utils.tool_utils import Tool as Tool
